<?php
namespace Codexonics\PrimeMoverFramework\classes;

/*
 * This file is part of the Codexonics.PrimeMoverFramework package.
 *
 * (c) Codexonics Ltd
 *
 * This package is Open Source Software. For the full copyright and license
 * information, please view the LICENSE file which was distributed with this
 * source code.
 */

use Codexonics\PrimeMoverFramework\utilities\PrimeMoverShutdownUtilities;

if (! defined('ABSPATH')) {
    exit;
}

/**
 * Prime Mover Error handling class
 *
 * The Prime Mover Error Handling Class handles the fatal runtime errors originated within the plugin and offers a systematic handling and reporting methods.
 *
 */
class PrimeMoverErrorHandlers
{
    
    private $system_functions;
    private $system_initialization;
    private $shutdown_utilities;
    private $system_authorization;
    
    /**
     *
     * Constructor
     */
    public function __construct(
        PrimeMoverSystemFunctions $system_functions, 
        PrimeMoverSystemInitialization $system_initialization,
        PrimeMoverShutdownUtilities $shutdown_utilities,
        PrimeMoverSystemAuthorization $system_authorization
        ) 
    {
            $this->system_functions = $system_functions;
            $this->system_initialization = $system_initialization;
            $this->shutdown_utilities = $shutdown_utilities;
            $this->system_authorization = $system_authorization;
    }

    /**
     * Get shutdown utilities
     * @return \Codexonics\PrimeMoverFramework\utilities\PrimeMoverShutdownUtilities
     * @compatible 5.6
     */
    public function getShutDownUtilities() 
    {
        return $this->shutdown_utilities;        
    }
    
    /**
     * Gets System authorization
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMoverSystemAuthorization
     * @compatible 5.6
     */
    public function getSystemAuthorization()
    {
        return $this->system_authorization;
    }
    
    /**
     *
     * Get system functions
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMoverSystemFunctions
     * @compatible 5.6
     */
    public function getSystemFunctions()
    {
        return $this->system_functions;
    }
    
    /**
     *
     * Get System Initialization
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMoverSystemInitialization
     * @compatible 5.6
     */
    public function getSystemInitialization()
    {
        return $this->system_initialization;
    }
    
    /**
     * Initialize error handling hooks
     * @compatible 5.6
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverErrorHandlers::itAddsInitHooks() 
     */
    public function initHooks()
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        add_action('wp_ajax_multisite_migration_check_if_error_log_exist', [$this,'multisiteMigrationLogExist']);
        
        add_action('shutdown', [$this,'errorHandler']);
        add_action('multisite_migration_shutdown_actions', [$this, 'multisiteMigrationDeleteMaintenanceFile'], 10, 1 );
        add_action('multisite_migration_shutdown_actions', [$this, 'multisiteMigrationDeletePackagePath'], 15 );
        
        add_action('multisite_migration_shutdown_actions', [$this, 'multisiteMigrationLogErrors'], 20, 1 );
        add_action('multisite_migration_before_doing_import', [$this, 'multisiteMigrationDeleteErrorLog'], 10, 1 );        
        
        add_action('init', [$this, 'streamError'], 1);
        add_action('multisite_migration_before_doing_export', [$this, 'multisiteMigrationDeleteErrorLog'], 10, 1 );
        add_action('multisite_migration_log_processed_events', [$this, 'multisiteMigrationLogEvents'], 10, 5);
        
        add_action('multisite_migration_bootup', [$this,'maybeRefreshLogEventFile'], 0, 3);
        add_filter('multisite_migration_filter_error_output', [$this, 'appendSiteIdentityOnErrorLog'], 10, 2);
        add_filter('multisite_migration_filter_error_output', [$this, 'appendCoreAndWordPressInfoOnLog'], 0, 1);
        
        add_filter('multisite_migration_filter_error_output', [$this, 'appendNetworkActivePluginsOnLog'], 100, 1);
        add_filter('multisite_migration_filter_error_output', [$this, 'appendPHPInfoArrayOnLog'], 300, 1);
    }
 
    /**
     * Append network active plugins on error logs
     * @param array $error_output
     * @return array
     */
    public function appendNetworkActivePluginsOnLog($error_output = [])
    {
        if ( ! is_array($error_output) ) {
            return $error_output;
        } 
        if ( ! is_multisite() ) {
            return $error_output;
        }
        $plugins = [];
        $network_active = $this->getSystemFunctions()->getSiteOption('active_sitewide_plugins');
        if ( ! is_array($network_active) || empty($network_active) ) {
            $error_output['active_sitewide_plugins'] = $plugins;
            return $error_output;
        }
        $keys = array_keys($network_active);
        clearstatcache();
        foreach ($keys as $plugin) {
            $plugin_full_path	= PRIME_MOVER_PLUGIN_CORE_PATH . $plugin;            
            if ( ! file_exists($plugin_full_path)) {
                continue;
            }            
            $plugins[] = get_plugin_data($plugin_full_path);
        }
        $error_output['active_sitewide_plugins'] = $plugins;
        return $error_output;
    }
    
    /**
     * Append core and WordPress info on log
     * @param array $error_output
     * @return array
     */
    public function appendPHPInfoArrayOnLog($error_output = [])
    {
        if ( ! is_array($error_output) ) {
            return $error_output;
        }
        
        $phpinfoarray = $this->getShutDownUtilities()->phpinfo2array();
        $error_output = array_merge($error_output, $phpinfoarray);        
        
        return $error_output;
    }
    
    /**
     * Append core and WordPress info on log
     * @param array $error_output
     * @return array
     */
    public function appendCoreAndWordPressInfoOnLog($error_output = [])
    {
        if ( ! is_array($error_output) ) {
            return $error_output;
        }
        
        $error_output['multisite_migration_version'] = PRIME_MOVER_VERSION;
        $error_output['wordpress_core_version'] = get_bloginfo('version');
        
        return $error_output;
    }
    
    /**
     * Append site identity on error log
     * @param array $error_output
     * @param number $blog_id
     * @return array
     */
    public function appendSiteIdentityOnErrorLog($error_output = [], $blog_id = 0)
    {
        if ( ! is_array($error_output) ) {
            return $error_output;
        }
        if ($blog_id) {
            $error_output['blog_id'] = $blog_id;
            $error_output['site_title'] = $this->getSystemFunctions()->getBlogOption($blog_id, 'blogname');
        }
        $error_output['super_admin_id'] = get_current_user_id();        
        return $error_output;
    }
    
    /**
     * Maybe refresh log event file
     * @param string $process_id
     * @param number $blog_id
     * @param boolean $diffmode
     */
    public function maybeRefreshLogEventFile($process_id = '', $blog_id = 0, $diffmode = false)
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $troubleshooting_path = $this->getSystemInitialization()->getTroubleShootingLogPath('migration'); 
        if ( ! $troubleshooting_path ) {
            return;
        }
        if ( ! $this->isLogPathValid($troubleshooting_path) ) {
            return;
        }
        $persist = false;
        if ( defined('PRIME_MOVER_LOG_PERSIST') && true === PRIME_MOVER_LOG_PERSIST ) {
            $persist = true;
        }
        $persist = apply_filters('multisite_migration_persist_troubleshooting_logs', $persist);
        clearstatcache();
        if (file_exists($troubleshooting_path) && ! $diffmode && ! $persist) {
            unlink($troubleshooting_path);
        }
    }
    
    /**
     * Log events as requested
     * @param $log
     * @param number $blog_id
     * @param string $mode
     * @param string $source
     * @param $object
     */
    public function multisiteMigrationLogEvents($log, $blog_id = 0, $mode = '', $source = '', $object)
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $troubleshooting_path = $this->getSystemInitialization()->getTroubleShootingLogPath('migration');        
        if ( ! $troubleshooting_path || ! $mode || ! $source  || ! is_object($object) ) {
            return;
        }
        if ( ! $this->isLogPathValid($troubleshooting_path) ) {
            return;
        }
        $disabled_by_default = true;
        if ($troubleshooting_path) {
            $disabled_by_default = false;
        }
        if (apply_filters('multisite_migration_disable_serverside_log', $disabled_by_default)) {
            return;
        }
        if (is_object($log)) {
            $log = (array)$log;
        }
        if (is_array($log)) {
            $log = $this->printError($log);
        }
        $time = date("Y-m-d H:i:s");
        $object_logged = get_class($object);
        $log = "$time => Logged $mode event for blog ID $blog_id from $object_logged::$source method: " . $log . PHP_EOL;
        $this->errorLog(apply_filters('multisite_migration_filter_error_log', $log, $source), $troubleshooting_path);
    }
    
    /**
     * Checks if log path is valid
     * @param string $path
     * @return boolean
     */
    protected function isLogPathValid($path = '')
    {
        if ( ! $path ) {
            return false;
        }
        $input = strtolower($path);
        return ('log' === pathinfo($input, PATHINFO_EXTENSION));
    }
    
    /**
     * Error log downloader
     * Hooked to `init`
     * @compatible 5.6
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverErrorHandlers::itStreamsErrorLog() 
     */
    public function streamError()
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        
        $params = $this->getShutDownUtilities()->getParameters();
        if (empty($params['multisite_migration_errornonce']) || empty($params['prime_mover_blogid'])) {
            return;
        }
        
        $blog_id = $params['prime_mover_blogid'];
        if (! wp_verify_nonce($params['multisite_migration_errornonce'], 'multisite_migration_errornonce' . $blog_id)) {
            return;
        }
        
        $mainsite_blog_id = $this->getSystemInitialization()->getMainSiteBlogId();
        $this->getSystemFunctions()->switchToBlog($mainsite_blog_id);
        $error_hash_option = $this->getSystemInitialization()->getErrorHashOption();
        
        $errorlogfile = $this->getShutDownUtilities()->getErrorHash($blog_id, $error_hash_option);
        $this->getSystemFunctions()->restoreCurrentBlog();
        
        if (file_exists($errorlogfile)) {
            $generatedFilename = $this->getSystemInitialization()->getErrorLogFile($blog_id);
            header('Content-Type: text/plain');
            header('Content-Disposition: attachment; filename="'. $generatedFilename .'"');
            header('Content-Length: ' . filesize($errorlogfile));
            
            flush();
            $this->getSystemFunctions()->readfileChunked($errorlogfile, false);
        }
        
        wp_die(0);
    }
    
    /**
     * AJAX handler for error log reporting
     * Hooked to `wp_ajax_multisite_migration_check_if_error_log_exist`
     * @compatible 5.6
     */
    public function multisiteMigrationLogExist()
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        
        //Initialize response
        $response = [];
        $response['logexist'] = false;
        
        //Initialize args
        $args = [
            'multisite_migration_errorlog_nonce' => FILTER_SANITIZE_STRING,
            'error_blog_id' => FILTER_SANITIZE_NUMBER_INT
        ];
        
        $error_post = $this->getSystemInitialization()->getUserInput('post', $args, 'errorlog_exist_check', '', 0, true);        
        if ( ! isset($error_post['error_blog_id']) ) {
            wp_die();
        }
        
        if ( ! isset($error_post['multisite_migration_errorlog_nonce'] ) ) {
            wp_die();
        }
        
        if ( ! wp_verify_nonce($error_post['multisite_migration_errorlog_nonce'], 'multisite_migration_errorlog_nonce')) {
            wp_die();
        }
        $blog_id = (int)$error_post['error_blog_id'];
        $error_log = $this->getSystemInitialization()->getErrorLogFile($blog_id);
        if ( $this->getShutDownUtilities()->multisiteMigrationErrorLogExist( false, $blog_id, $error_log) ) {
            $errorlog_url = $this->getShutDownUtilities()->getDownloadErrorLogURL( $blog_id );
            
            $response['logexist'] = true;
            $response['error_msg'] = esc_html__('Runtime Error : ', 'prime-mover' ) . '  ' . '<a href="' . esc_url($errorlog_url) . '">' .
                esc_html__( 'View error', 'prime-mover') . '</a>';
        }
        
        //Update client of the status        
        wp_send_json($response);
    }
    
    /**
     * Multisite migration log errors
     * Designed to work even WP_DEBUG is false
     * Hooked to `multisite_migration_shutdown_actions`
     * @param array $error
     * @compatible 5.6
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverErrorHandlers::itLogsPrimeMoverErrors()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverErrorHandlers::itDoesNotLogMigrationErrorsIfNotAuthorized()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverErrorHandlers::itDoesNotLogErrorIfNoError()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverErrorHandlers::itDoesNotLogErrorIfIdIsNotSet()
     */
    public function multisiteMigrationLogErrors(array $error)
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        
        if ( ! is_array($error) || empty($error) ) {
            return;
        }
        $blog_id = $this->getShutDownUtilities()->multisiteMigrationGetProcessedID();
        if ( ! $blog_id ) {
            return;
        }
        global $wp_filesystem;
        $error_log = $this->getSystemInitialization()->getErrorLogFile($blog_id);
        $error_log_file = $this->getShutDownUtilities()->getPrimeMoverErrorPath($blog_id, $error_log);
        $report_instructions = '### You can report this issue to https://codexonics.com/ for tips and bug fixing ###';
        $report_instructions .= PHP_EOL;
        
        if ( ! $wp_filesystem->exists($error_log_file)) {
            $wp_filesystem->put_contents($error_log_file, $report_instructions, FS_CHMOD_FILE);
        }

        $error_output = apply_filters('multisite_migration_filter_error_output', $error, $blog_id);
        
        $error_string = $this->printError($error_output); 
        $error_hash_option = $this->getSystemInitialization()->getErrorHashOption();
        
        $this->getShutDownUtilities()->logErrorHash($blog_id, $error_log_file, $error_hash_option);
        $this->errorLog($error_string, $error_log_file);
    }
    
    /**
     * Print error
     * @param array $error_output
     */
    protected function printError($error_output = [])
    {
        return print_r($error_output,true);
    }
    
    /**
     * Error log
     * @param string $error_string
     * @param string $error_log_file
     */
    protected function errorLog($error_string = '', $error_log_file = '')
    {
        @error_log($error_string, 3, $error_log_file);
    }
    
    /**
     * Delete error log before starting any processes
     * Hooked to `multisite_migration_before_doing_import`
     * Hooked to `multisite_migration_before_doing_export`
     * @param int $blog_id
     * @compatible 5.6
     */
    public function multisiteMigrationDeleteErrorLog($blog_id = 0)
    {        
        $this->getShutDownUtilities()->multisiteMigrationDeleteErrorLog($blog_id);
    }
    
    /**
     * Deletes temporary packages in the event of runtime errors to prevent clogging up the server
     * Hooked to `multisite_migration_shutdown_actions`
     * @compatible 5.6
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverErrorHandlers::itDeletesPackageOnFatalError() 
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverErrorHandlers::itDoesNotDeletePackageUserCreatedPackage() 
     */
    public function multisiteMigrationDeletePackagePath()
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        global $wp_filesystem;
        $processed_path = $this->getShutDownUtilities()->multisiteMigrationGetProcessedPath();
        $blog_id = $this->getShutDownUtilities()->multisiteMigrationGetProcessedID();
        if ( is_array( $processed_path ) ) {
            foreach ( $processed_path as $path_to_delete ) {
                if ( $wp_filesystem->exists($path_to_delete) && apply_filters('multisite_migration_delete_package_on_error', true, $path_to_delete, $blog_id)) {
                    $this->getSystemFunctions()->multisiteMigrationDoDelete($path_to_delete);
                }
            }
        }
    }
    
    /**
     * Multisite migration error handlers on shutdown
     * This fires successfully when a fatal error is detected inside Multisite migration plugin
     * Hooked to `shutdown`
     * @compatible 5.6
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverErrorHandlers::itRunsErrorHandlerWhenItsFatalError() 
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverErrorHandlers::itDoesNotRunErrorHandlerWhenErrorIsNotFatal()
     */
    public function errorHandler()
    {
        if ( ! is_admin() ) {
            return;
        }
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $error = $this->multisiteMigrationGetErrorLast();
        if( is_null( $error ) ) {
            return;
        }
        if ( ! is_array( $error ) || empty( $error['file'] ) || empty( $error['type'] ) ) {
            return;
        }
        if (E_ERROR !== $error['type']) {
            return;
        }        
        if (function_exists('wp_raise_memory_limit')) {
            wp_raise_memory_limit();
        }        
        do_action( 'multisite_migration_shutdown_actions', $error );
    }
    
    /**
     * Get error last
     * @return array
     */
    protected function multisiteMigrationGetErrorLast()
    {
        return error_get_last();
    }
    
    /**
     /* Multisite migration deletes maintenance file
      * Hooked to `multisite_migration_shutdown_actions`
      * @param array $error
      * @compatible 5.6
     */
    public function multisiteMigrationDeleteMaintenanceFile(array $error) {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        if ( ! $this->getShutDownUtilities()->maybeDeleteMaintenanceFile( $error['type'] ) ) {
            return;
        }
        $this->getSystemFunctions()->disableMaintenanceDuringImport();
    }
}