<?php
namespace Codexonics\PrimeMoverFramework\classes;

/*
 * This file is part of the Codexonics.PrimeMoverFramework package.
 *
 * (c) Codexonics Ltd
 *
 * This package is Open Source Software. For the full copyright and license
 * information, please view the LICENSE file which was distributed with this
 * source code.
 */

if (! defined('ABSPATH')) {
    exit;
}

/**
 * Prime Mover Hooked Methods
 *
 * The Prime Mover Hooked Method Class provides the hooked methods to WordPress core hooks
 *
 */
class PrimeMoverHookedMethods
{
    private $system_authorization;
    private $system_initialization;
    private $system_checks;
    private $system_functions;
    private $progress_handlers;
    
    /**
     * Constructor
     * @param PrimeMoverSystemAuthorization $system_authorization
     * @param PrimeMoverSystemInitialization $system_initialization
     * @param PrimeMoverSystemChecks $system_checks
     * @param PrimeMoverSystemFunctions $system_functions
     * @param PrimeMoverProgressHandlers $progress_handlers
     */
    public function __construct(
        PrimeMoverSystemAuthorization $system_authorization,
        PrimeMoverSystemInitialization $system_initialization,
        PrimeMoverSystemChecks $system_checks,
        PrimeMoverSystemFunctions $system_functions,
        PrimeMoverProgressHandlers $progress_handlers
    ) 
    {
        $this->system_authorization = $system_authorization;
        $this->system_initialization = $system_initialization;
        $this->system_checks = $system_checks;
        $this->system_functions = $system_functions;
        $this->progress_handlers = $progress_handlers;
    }

    /**
     * Get progress handlers
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMoverProgressHandlers
     */
    public function getProgressHandlers()
    {
        return $this->progress_handlers;        
    }
    
    /**
     * Get system functions
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMoverSystemFunctions
     * @compatible 5.6
     */
    public function getSystemFunctions()
    {
        return $this->system_functions;
    }
    
    /**
     * Get System authorization
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMoverSystemAuthorization
     * @compatible 5.6
     */
    public function getSystemAuthorization()
    {
        return $this->system_authorization;
    }
    
    /**
     * Get System checks
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMoverSystemChecks
     * @compatible 5.6
     */
    public function getSystemChecks()
    {
        return $this->system_checks;
    }
    
    /**
     * Get System Initialization
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMoverSystemInitialization
     * @compatible 5.6
     */
    public function getSystemInitialization()
    {
        return $this->system_initialization;
    }
    
    /**
     * Adds the custom export and import column to Network -> Sites list table.
     * @param array $column_headers
     * @return array $column_headers
     * @compatible 5.6
     */
    public function multisiteMigrationAddNetworkColumn($column_headers = [])
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        /**
         * Add custom export/import columns if export folder created
         */
        if (true === $this->getSystemChecks()->multisiteMigrationEssentialRequisites()) {
            $column_headers['multisite_export_column']	=	esc_html__('Export Site', 'prime-mover');
            $column_headers['multisite_import_column'] 	=	esc_html__('Import Site', 'prime-mover');
        }
        
        return $column_headers ;
    }
    
    /**
     * Get blog address
     * @param number $blog_id
     * @return string
     */
    private function getBlogAddress($blog_id = 0)
    {
        if (is_multisite()) {
            return get_blogaddress_by_id($blog_id);
        } else {
            return network_site_url();
        }        
    }
    
    /**
     * Maybe load export section check
     * @param string $column_name
     * @param number $blog_id
     * @return boolean
     */
    private function maybeLoadMigrationSection($column_name = '', $blog_id = 0, $mode = 'export')
    {
        $columns = ['export' => 'multisite_export_column', 'import' => 'multisite_import_column'];
        if ( ! isset($columns[$mode] ) ) {
            return false;
        }
        $column_check = $columns[$mode];
        if (is_multisite() && ($column_check === $column_name) && ($blog_id > 0) && ($blog_id != 1)) {
            return true;
        }
        if ( ! is_multisite() ) {
            return true;
        }
        return false;
    }
    
    /**
     * Renders the export section of the plugin in Network -> Sites
     * @param string $column_name
     * @param number $blog_id
     * @compatible 5.6
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverHookedMethods::itOutputsPrimeMoverExportSection()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverHookedMethods::itDoesNotOutputExportSectionNotOnExportColumn()
     */
    public function multisiteMigrationExportSection($column_name = '', $blog_id = 0)
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        if ( ! is_multisite() && ! $blog_id ) {
            $blog_id = 1;
        }

        /**
         * Only proceed if export folder created
         */
        if (true === $this->getSystemChecks()->multisiteMigrationEssentialRequisites()) {
            $button_text = $this->generateExportButtonText($blog_id);
            $blog_id	= intval($blog_id);
            if ($this->maybeLoadMigrationSection($column_name, $blog_id)) {
                $blogaddress_by_id	= $this->getBlogAddress($blog_id);
                $tooltip_button = sprintf(esc_html__('Export %s with blog ID: %d', 'prime-mover'), $blogaddress_by_id, $blog_id);
                if ( ! is_multisite() ) {
                    $tooltip_button = esc_html__('Export site', 'prime-mover');
                }
                ?>
			<input	name="multisite_migration_exportbutton" 
					value="<?php echo $button_text; ?>" 
					class="button-primary multisite_migration_exportbutton js-multisite_migration_exportbutton"
					title="<?php echo esc_attr($tooltip_button); ?>"
					id	 ="js-multisite_migration_exporting_blog_<?php echo esc_attr($blog_id) ; ?>"
					data-multisiteblogid	= "<?php echo esc_attr($blog_id); ?>" 				
					type="button" >				
			<p class="multisite_migration_export_progress_span_p" id="js-multisite_migration_export_progress_span_p_<?php echo esc_attr($blog_id); ?>">
			   <span id="js-multisite_export_span_<?php echo esc_attr($blog_id); ?>"></span>
			   <span class="multisite_migration_progress_span" id="js-multisite_export_progress_span_<?php echo esc_attr($blog_id); ?>"></span>
			</p>
			<?php do_action('multisite_migration_do_after_export_button', $blog_id); ?>
			<?php
            }
        }
    }
    
    /**
     * Generate export button text
     * @param number $blog_id
     * @return string
     */
    private function generateExportButtonText($blog_id = 0)
    {
        $text = sprintf(('%s : %d'), esc_html__('Export blog ID', 'prime-mover'), esc_attr($blog_id));
        if (wp_is_mobile() || ! is_multisite() ) {
            $text = esc_html__('Export', 'prime-mover');
        }
        
        return $text;       
    }
    
    /**
     * Renders the export section of the plugin in Network -> Sites
     * @param string $column_name
     * @param number $blog_id
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverHookedMethods::itOutputsPrimeMoverImportSection()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverHookedMethods::itDoesNotOutputImportSectionIfNotImportColumn()
     * @compatible 5.6
     */
    public function multisiteMigrationImportSection($column_name = '', $blog_id = 0)
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        if ( ! is_multisite() && ! $blog_id ) {
            $blog_id = 1;
        }
        if (true === $this->getSystemChecks()->multisiteMigrationEssentialRequisites()) {
            $blog_id	= intval($blog_id);
            if ($this->maybeLoadMigrationSection($column_name, $blog_id, 'import')) {
                ?>
            <?php do_action('multisite_migration_do_before_import_button', $blog_id); ?>
	    	<?php do_action('multisite_migration_render_import_button', $blog_id ); ?> 
	    	<p class="multisite_migration_import_progress_span_p" id="js-multisite_migration_import_progress_span_p_<?php echo esc_attr($blog_id); ?>">
	    		<span class="multisite_migration_import_span" id="js-multisite_import_span_<?php echo esc_attr($blog_id); ?>"></span>
	    		<span class="multisite_migration_progress_span" id="js-multisite_import_progress_span_<?php echo esc_attr($blog_id); ?>"></span>
	    	</p>
	    	<?php do_action('multisite_migration_do_after_import_button', $blog_id); ?>		    
	<?php
            }
        }
    }

    /**
     * Add js body class
     * @param string $classes
     * @return string
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverHookedMethods::itAddsJsBodyClassOnNetworkSitesPage()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverHookedMethods::itDoesNotAddBodyClassNotOnNetworkSites()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverHookedMethods::itAddsJsBodyClassOnExistingBodyClass() 
     */
    public function addJsBodyClassOnNetworkSitesPage($classes = '')
    {       
        if ( ! $this->getSystemInitialization()->isNetworkSites()) {
            return $classes;
        }
        if (empty($classes)) {
            $classes_array = [];
        } else {
            $classes_array = explode(' ', $classes);
        }        
        $js_body_class = $this->getSystemInitialization()->getJsBodyClass();
        
        if ( ! in_array($js_body_class, $classes_array)) {
            $classes_array[] = $js_body_class;
        }
        $classes = implode(' ', $classes_array);
        return $classes;
    }
    
    /**
     * Checks if we are not on network sites page
     * @param string $hook
     * @return boolean
     */
    private function maybeWeAreNotOnSitesPage($hook = '')
    {
        global $current_screen;
        if (is_multisite()) {
            return ('sites.php' != $hook || ! $current_screen->in_admin('network') || false === $this->getSystemChecks()->multisiteMigrationEssentialRequisites());
        }
        if ( ! is_multisite() ) {
            $load_hook = 'tools_page_migration-tools';            
            return ($load_hook != $hook || $current_screen->id !== $load_hook || false === $this->getSystemChecks()->multisiteMigrationEssentialRequisites());
        }
    }
    
    /**
     * Enqueue plugin JS on Network -> Sites.
     * @param string $hook
     * @compatible 5.6
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverHookedMethods::itEnqueuesStandardPrimeMoverScripts() 
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverHookedMethods::itEnqueuesMinifiedJsOnNonDebugMode() 
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverHookedMethods::itDoesNotEnqueueScriptNotOnSitesPage()
     */
    public function multisiteMigrationEnqueueScripts($hook = '')
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        
        if ($this->getSystemFunctions()->maybeLoadMenuAssets()) {
            wp_enqueue_style(
                'multisite_migration_panel_dashicons',
                esc_url_raw(plugins_url('res/css/prime-mover-panel-icon.css', dirname(__FILE__))),
                [],
                PRIME_MOVER_VERSION
                );
        }
        
        if ($this->maybeWeAreNotOnSitesPage($hook)) {
            return;
        }
        $this->getSystemInitialization()->setIsNetworkSites(true);
        $min = '.min';
        if ( defined('SCRIPT_DEBUG') && SCRIPT_DEBUG ) {
            $min = '';
        }
        $core_js = "prime-mover-js-network-admin$min.js";
        $jszip = "zip.js";
        
        if ( ! is_multisite() ) {
            remove_all_actions( 'admin_notices' );
        }
        
        wp_enqueue_script(
            'multisite_migration_js_network_admin',
            esc_url_raw(plugins_url('res/js/' . $core_js, dirname(__FILE__))),
            ['jquery', 'jquery-ui-core', 'jquery-ui-dialog'],
                PRIME_MOVER_VERSION
        );
        
        wp_enqueue_script(
            'multisite_migration_jszip',
            esc_url_raw(plugins_url('res/js/' . $jszip, dirname(__FILE__))),
            ['jquery', 'jquery-ui-core', 'jquery-ui-dialog', 'multisite_migration_js_network_admin'],
            PRIME_MOVER_VERSION
            );
        
        wp_enqueue_style(
            'multisite_migration_css_network_admin',
                esc_url_raw(plugins_url('res/css/prime_mover_css_network_admin.css', dirname(__FILE__))),
                ['wp-jquery-ui-dialog'],
                PRIME_MOVER_VERSION
                );
        
        $debug_uploads = false;
        if (defined('PRIME_MOVER_DEBUG_UPLOAD') && PRIME_MOVER_DEBUG_UPLOAD) {
            $debug_uploads = true;
        }
        wp_localize_script(
            'multisite_migration_js_network_admin',
            'multisite_migration_js_ajax_renderer',
                apply_filters( 'multisite_migration_ajax_rendered_js_object', [
                        'ajaxurl' => esc_url_raw(admin_url('admin-ajax.php')),
                        'multisite_migration_slice_size' => $this->getSystemFunctions()->getSliceSize(),
                        'multisite_migration_export_nonce' => wp_create_nonce('multisite_migration_export_nonce'),
                        'multisite_migration_import_nonce' => wp_create_nonce('multisite_migration_import_nonce'),
                        'multisite_migration_uploads_nonce' => wp_create_nonce('multisite_migration_uploads_nonce'),
                        'multisite_migration_import_progress_nonce' => wp_create_nonce('multisite_migration_import_progress_nonce'),
                        'multisite_migration_export_progress_nonce' => wp_create_nonce('multisite_migration_export_progress_nonce'),
                        'prime_mover_decryption_check_nonce' => wp_create_nonce('prime_mover_decryption_check_nonce'),
                        'multisite_migration_errorlog_nonce' => wp_create_nonce('multisite_migration_errorlog_nonce'),
                        'multisite_migration_deletetmpfile_nonce' => wp_create_nonce('multisite_migration_deletetmpfile_nonce'),
                        'multisite_migration_upload_progress' => esc_js(__('Uploading in progress.', 'prime-mover')),
                        'multisite_migration_ajax_ajax_loader_gif' => esc_url_raw(plugins_url('res/img/ajax-loader.gif', dirname(__FILE__))),
                        'multisite_migration_import_success' => esc_js(__('Site successfully imported!', 'prime-mover')),
                        'multisite_migration_import_error' => esc_js(__('Import error!', 'prime-mover')),
                        'export_now_button' => esc_js(__('Export now', 'prime-mover')),
                        'cancel_button' => esc_js(__('Cancel', 'prime-mover')),
                        'yes_button' => esc_js(__('Yes', 'prime-mover')),
                        'no_button' => esc_js(__('No', 'prime-mover')),
                        'ok_button' => esc_js(__('OK', 'prime-mover')),
                        'multisite_migration_post_max_size' => $this->getSystemFunctions()->getPostMaxSizeInPhpini(),
                        'multisite_migration_upload_max_size' => $this->getSystemFunctions()->getUploadmaxFilesizeInPhpini(),
                        'multisite_migration_debug_uploads' => apply_filters('multisite_migration_enable_upload_js_debug', $debug_uploads),
                        'multisite_migration_export_downloaded' => esc_js(__('Export downloaded to computer', 'prime-mover')),
                        'multisite_migration_spinner_upload_text' => esc_js(__('Processing', 'prime-mover')),
                        'multisite_migration_spinner_download_text' => esc_js(__('Initializing download', 'prime-mover')),
                        'multisite_migration_downloading_progress_text' => esc_js(__('Downloading package', 'prime-mover')),
                        'multisite_migration_dropbox_upload_progress_text' => esc_js(__('Uploading to Dropbox', 'prime-mover')),
                        'multisite_migration_unknown_js_error' => esc_js(__('Unknown error occurred, please try again.', 'prime-mover')),
                        'multisite_migration_spinner_zipanalysis_text' => esc_js(__('Analyzing zip..', 'prime-mover')),
                        'prime_mover_zipjs_workers' => esc_url_raw(plugins_url('res/js/zip-library/', dirname(__FILE__))),
                        'prime_mover_media_decryption_error' => $this->getSystemInitialization()->returnCommonMediaDecryptionError(),
                        'prime_mover_phpuploads_misconfigured' => $this->getSystemFunctions()->maybeUploadParametersMisconfigured(),
                        'prime_mover_upload_misconfiguration_error' => esc_js(__('Aborting upload ! PHP upload is misconfigured. POST_MAX_SIZE setting in php.ini should be greater than UPLOAD_MAX_FILESIZE', 'prime-mover')),
                        'prime_mover_invalid_package' => esc_js(__('Invalid file type or corrupted zip file.', 'prime-mover')),
                        'prime_mover_exceeded_browser_limit' => esc_js(__('Restoring package beyond 4GB is not supported by browser uploads. Please upgrade to premium version and use remote URL restore feature.', 'prime-mover')),
                ])
        );
        
        do_action( 'multisite_migration_after_enqueue_assets');
    }

    /**
     * Show an update nag to network administrator
     * that the plugin was not been able to initialize due to unmeet dependencies
     * @compatible 5.6
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverHookedMethods::itShowsAdminNoticeErrorWhenNoOpenSSLSupport()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverHookedMethods::itDoesNotOutputAnyErrorWhenAllRequisitesMeet()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverHookedMethods::itReturnsErrorWhenFolderIsNotCreated()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverHookedMethods::itReturnsErrorWhenFileSystemIsNotInitialized()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverHookedMethods::itReturnsErrorWhenZipDisabled()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverHookedMethods::itReturnsErrorWhenNotCompliedWithMinimumRequirement()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverHookedMethods::itReturnsErrorWhenNotAllSystemsFunctionsEnabled() 
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverHookedMethods::itReturnsErrorWhenNoMbStringEnabled()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverHookedMethods::itReturnsErrorWhenFileInfoDisabled
     */
    public function multisiteShowNetworkAdminNotice()
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        if (false === $this->getSystemChecks()->multisiteMigrationEssentialRequisites()) {
            //Essential requirements not meet ?>
			<div class="error">
		<?php 
            
        if (false === $this->getSystemInitialization()->getMultisiteExportFolderCreated()) {
            ?>				
			<p><?php echo esc_html__('Prime Mover plugin is not yet ready to use. Error: Unable to create its own export folder. 
					This is required to export sites. Please make sure WordPress has permission to create folder in your uploads directory.', 'prime-mover'); ?></p>
				
			<?php
        }
            if (false === $this->getSystemInitialization()->getMultisiteWpFilesystemInitialized()) {
                ?>			
			<p><?php echo sprintf ( esc_html__('Prime Mover plugin is activated but not yet ready to use. %s.', 
			    'prime-mover'), 
			    '<strong>' . esc_html__('Error: WordPress FileSystem API not set since it requires DIRECT FILE PERMISSIONS', 'prime-mover') . '</strong>'); ?></p>
			<p><?php echo esc_html__('Make sure WordPress is creating files as the same owner as the WordPress files.', 'prime-mover'); ?></p>
			
		<?php
            }
            if (false === $this->getSystemFunctions()->multisiteMigrationCheckIfZipEnabled()) {
                ?>
			<p><?php echo esc_html__('Prime Mover plugin is not yet ready to use. Error: Requires PHP Zip extension to be enabled. Please check with you web host.', 'prime-mover'); ?></p>			
		<?php
            }
            if (false === $this->getSystemFunctions()->compliedMinimumRequirement()) {
                //Essential PHP version requirements not meet ?>
			<p><?php echo esc_html__('Prime Mover plugin is not yet ready to use. Error: Requires at least PHP 5.6.0.', 'prime-mover'); ?></p>
		<?php
            }
            if (false === $this->getSystemChecks()->multisiteMigrationCheckIfMbstringEnabled()) {
                //No mbstring extension ?>
			<p><?php echo esc_html__('Prime Mover plugin is not yet ready to use. Error: Requires mbstring PHP extension. Please enable it.', 'prime-mover'); ?></p>
		<?php
            }
            if (false === $this->getSystemChecks()->multisiteMigrationCheckIfFileinfoExtensionEnabled()) {
                //No fileinfo extension ?>
			<p><?php echo esc_html__('Prime Mover plugin is not yet ready to use. Error: Requires fileinfo PHP extension. Please enable it.', 'prime-mover'); ?></p>
		<?php
            }
        ?>
		</div>
		<?php
        }
    }
    
    /**
     * Single site migration callback
     */
    public function singleSiteMigrationCallBack()
    {
    ?>
        <div class="wrap">
           <h1><?php esc_html_e( 'Migration Tools', 'prime-mover' ); ?></h1>
           <div class="card">              
                 <div class="multisite_migration_exporter_block">  
                 <h2><?php esc_html_e( 'Export / Backup', 'prime-mover' ); ?></h2>             
                     <div class="notice-large highlight multisite-migration-helper-export-text">
                         <p>
                         <?php esc_html_e( 'This export tool can be used to migrate WordPress to the following supported modes', 'prime-mover' );?> :
                         </p> 
                        <ul class="ul-square">
                            <li><?php esc_html_e( 'Migrate single-site to another server single-site installation and vice versa', 'prime-mover' );?>.</li>
                            <li><?php esc_html_e( 'Migrate single-site to a multisite subsite installation and vice versa', 'prime-mover' );?>.</li>
                            <li><?php esc_html_e( 'Backup single site so it can be restored anytime', 'prime-mover' );?>.</li>
                        </ul>
                     </div>                                      
                     <?php                        
                         do_action('multisite_migration_exporter_block');
                     ?>
                 </div>
          </div>    
          <div class="card">             
                 <div class="multisite_migration_importer_block">    
                 <h2><?php esc_html_e( 'Import / Restore', 'prime-mover' ); ?></h2> 
                     <div class="notice-large highlight multisite-migration-helper-export-text">
                     <?php esc_html_e( 'This import tool can be used to restore an export package with the following restore modes ', 'prime-mover' ); 
                     ?> :                  
                         <ul class="ul-square">
                            <li><?php esc_html_e( 'Restore single-site package from another site to this site and vice versa', 'prime-mover' );?>.</li>
                            <li><?php esc_html_e( 'Restore single-site package from multisite subsite to this site and vice versa', 'prime-mover' );?>.</li>
                            <li><?php esc_html_e( 'Restore a single-site backup created by this site and vice versa', 'prime-mover' );?>.</li>
                        </ul>
                        <p><em><?php echo esc_html__( 'Please do not refresh or navigate away from this page while import is ongoing.', 
                             'prime-mover' ); 
                         ?></em></p>                                               
                     </div>                                                                        
                      <?php 
                         do_action('multisite_migration_importer_block');
                      ?>
                </div>       
	       </div> 
	   </div>
   <?php      
    }
    
    /**
     * Callback for single site Tools sub-menu
     */
    public function addExportImportOptionsSingleSite()
    {
        add_submenu_page( 'tools.php', esc_html__('Migration Tools', 'prime-mover'), 
            esc_html__('Migration Tools', 'prime-mover'),
            'manage_options', 'migration-tools', [$this, 'singleSiteMigrationCallBack']);
    }
    
    /**
     * Add menu page
     */
    public function addMenuPage()
    {
        if ( ! $this->getSystemFunctions()->maybeLoadMenuAssets() ) {
            return;
        }
        $required_cap = 'manage_network_options';
        if ( ! is_multisite() ) {
            $required_cap = 'manage_options';
        }
        
        add_menu_page( esc_html__('Prime Mover Control Panel', 'prime-mover'), esc_html__('Prime Mover', 'prime-mover'),
        $required_cap, 'migration-panel-settings', [$this, 'addMenuPageCallBack'], 'dashicons-multisitemigrationpaneldashicons');
        
        do_action('prime_mover_run_menus');
    }
    
    /**
     * Add menu page callback
     */
    public function addMenuPageCallBack()
    {
        ?>
      <div class="wrap">
         <h1><?php esc_html_e('Prime Mover Control Panel', 'prime-mover')?></h1>
           <?php do_action('prime_mover_dashboard_content');?> 
      </div>       
    <?php     
    }
    
    /**
     * Remove distractions on Getting Started / Settings page
     */
    public function removeDistractionsOnSettingsPage()
    {
        if ( ! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $current_screen = get_current_screen();
        
        if ( $this->getSystemFunctions()->maybeLoadAssetsOnDashboard($current_screen ) ) {
            remove_all_actions( 'admin_notices' );
        }
    }
    
    /**
     * Load plugin text domain
     */
    public function loadPluginTextdomain()
    {
        load_plugin_textdomain( 'prime-mover', false, basename(PRIME_MOVER_PLUGIN_PATH) . '/languages/' );
    }
}
