<?php
namespace Codexonics\PrimeMoverFramework\classes;

/*
 * This file is part of the Codexonics.PrimeMoverFramework package.
 *
 * (c) Codexonics Ltd
 *
 * This package is Open Source Software. For the full copyright and license
 * information, please view the LICENSE file which was distributed with this
 * source code.
 */

use Codexonics\PrimeMoverFramework\interfaces\PrimeMoverImport;
use Codexonics\PrimeMoverFramework\utilities\PrimeMoverSearchReplace;

if (! defined('ABSPATH')) {
    exit;
}

/**
 * The Prime Mover Import Class
 *
 * The Prime Mover Import Class aims to provide the import facility of this plugin.
 *
 */
class PrimeMoverImporter implements PrimeMoverImport
{
    private $system_authorization;
    private $system_functions;
    private $system_initialization;
    private $system_checks;
    private $progress_handlers;
    
    /**
     * Constructor
     * @param PrimeMoverSystemAuthorization $system_authorization
     * @param PrimeMoverSystemFunctions $system_functions
     * @param PrimeMoverSystemInitialization $system_initialization
     * @param PrimeMoverSystemChecks $system_checks
     */
    public function __construct(
        PrimeMoverSystemAuthorization $system_authorization,
        PrimeMoverSystemFunctions $system_functions,
        PrimeMoverSystemInitialization $system_initialization,
        PrimeMoverSystemChecks $system_checks,
        PrimeMoverProgressHandlers $progress_handlers
    ) 
    {
        $this->system_authorization = $system_authorization;
        $this->system_functions = $system_functions;
        $this->system_initialization = $system_initialization;
        $this->system_checks = $system_checks;
        $this->progress_handlers = $progress_handlers;        
    }
    
    /**
     * Get progress handlers
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMoverProgressHandlers
     * @compatible 5.6
     */
    public function getProgressHandlers() 
    {
        return $this->progress_handlers;
    }
    
    /**
     * Get System checks
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMoverSystemChecks
     * @compatible 5.6
     */
    public function getSystemChecks()
    {
        return $this->system_checks;
    }
    
    /**
     * Get system functions
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMoverSystemFunctions
     * @compatible 5.6
     */
    public function getSystemFunctions()
    {
        return $this->system_functions;
    }
 
    /**
     * Get System Initialization
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMoverSystemInitialization
     * @compatible 5.6
     */
    public function getSystemInitialization()
    {
        return $this->system_initialization;
    }
    
    /**
     * Gets System authorization
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMoverSystemAuthorization
     * @compatible 5.6
     */
    public function getSystemAuthorization()
    {
        return $this->system_authorization;
    }
    
    /**
     * Importer hooks
     * @compatible 5.6
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverImporter::itAddsImporterHooks()
     */
    public function importerHooks()
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        /**
         * Pre-processing filters
         */
        add_filter('multisite_migration_filter_pre_comparison_package_array', [$this, 'filterPreComparisonPackageArrayFunc'], 10, 3);
        add_filter('multisite_migration_filter_pre_comparison_target_array', [$this, 'filterPreComparisonTargetArrayFunc'], 10, 3);
        
        add_filter('multisite_migration_filter_pre_comparison_target_array', [$this, 'removeSchemeInPrefilterComparison'], 10, 3);
        add_filter('multisite_migration_filter_pre_comparison_package_array', [$this, 'removeSchemeInPrefilterComparison'], 10, 3);
        
        add_filter('multisite_migration_filter_pre_comparison_target_array', [$this, 'removeUploadinfoInPrefilterComparison'], 10, 3);
        add_filter('multisite_migration_filter_pre_comparison_package_array', [$this, 'removeUploadinfoInPrefilterComparison'], 10, 3);
        
        //Step1. Move files from $_FILES to uploads temporary directory
        add_filter('multisite_migration_do_import', [$this, 'moveImportedFilesToUploads'], 0, 3);
        
        //Step2. Unzip
        add_filter('multisite_migration_do_import', [$this, 'unzipImportedZipPackageMigration'], 10, 3);
        
        //Step3. Validate if imported site matches with the package
        add_filter('multisite_migration_do_import', [$this, 'validateImportedSiteVsPackage'], 20, 3);
        
        //Step4. Compare system footprint of the origin site vs the imported package
        add_filter('multisite_migration_do_import', [$this, 'compareSystemFootprintImport'], 30, 3);
        
        //Step5. Update media files
        add_filter('multisite_migration_do_import', [$this, 'updateTargetMediaFilesWithNew'], 40, 3);

        //Step6. Mark target site uploads information
        add_filter('multisite_migration_do_import', [$this, 'markTargetSiteUploadsInformation'], 50, 3);
        
        //Step7. Optionally import plugins and themes
        add_filter('multisite_migration_do_import', [$this, 'multisiteOptionallyImportPluginsThemes'], 60, 3);
                
        //Step8. Import new dB
        add_filter('multisite_migration_do_import', [$this, 'importDb'], 80, 3);

        //Step9. Rename db_prefix
        add_filter('multisite_migration_do_import', [$this, 'renameDbPrefix'], 90, 3);
        
        //Step10. Search and replace
        add_filter('multisite_migration_do_import', [$this, 'searchAndReplace'], 100, 3);
        
        //Step11. Activate plugins if not still activated (e.g. if they are only network activated in source site)
        add_filter('multisite_migration_do_import', [$this, 'activatePluginsIfNotActivated'], 110, 3);
        
        //Step12. Handle uploads directory specific adjustments
        add_filter('multisite_migration_do_import', [$this, 'restoreCurrentUploadsInformation'], 120, 3);
        
        //Step13. Ends the import
        add_filter('multisite_migration_do_import', [$this, 'markImportSuccess'], 130, 3);
    }
    
    /**
     * Optionally import plugins and themes
     * {@inheritDoc}
     * @see \Codexonics\PrimeMoverFramework\interfaces\PrimeMoverImport::multisiteOptionallyImportPluginsThemes()
     * @compatible 5.6
     */
    public function multisiteOptionallyImportPluginsThemes($ret = [], $blogid_to_import = 0, $files_array = [])
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $this->getSystemInitialization()->setSlowProcess();
        if ((isset($ret['error'])) || (isset($ret['diff']))) {
            //Errors or diff still detected, bail out and inform user
            $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Errors or diffs detected, alert user.', 'prime-mover'));
            return $ret;
        }
        
        //All set. Proceed.
        //Import plugins
        do_action('multisite_migration_import_plugins', $ret, $blogid_to_import);
        
        //Import themes
        do_action('multisite_migration_import_themes', $ret, $blogid_to_import);
        
        do_action('multisite_migration_log_processed_events', $ret, $blogid_to_import, 'import', 'multisiteOptionallyImportPluginsThemes', $this);
        
        return $ret;
    }

    /**
     * Remove upload info in prefilter_comparison
     * @param array $system_footprint_site
     * @param array $ret
     * @param int $blogid_to_import
     * @return array
     * @compatible 5.6
     */
    public function removeUploadinfoInPrefilterComparison(array $system_footprint_site, array $ret, $blogid_to_import = 0)
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        
        if (isset($system_footprint_site['upload_information_path'])) {
            unset($system_footprint_site['upload_information_path']);
        }
        
        if (isset($system_footprint_site['upload_information_url'])) {
            unset($system_footprint_site['upload_information_url']);
        }
        
        return $system_footprint_site;
    }
    
    /**
     * Remove scheme in prefilter_comparison
     * @param array $system_footprint_site
     * @param array $ret
     * @param int $blogid_to_import
     * @return array
     * @compatible 5.6
     */
    public function removeSchemeInPrefilterComparison(array $system_footprint_site, array $ret, $blogid_to_import = 0)
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        
        if (isset($system_footprint_site['scheme'])) {
            unset($system_footprint_site['scheme']);
        }
        
        return $system_footprint_site;
    }
    
    /**
     * Move uploaded files to temporary uploads directory
     * {@inheritDoc}
     * @see PrimeMoverImport::moveImportedFilesToUploads()
     * @compatible 5.6
     */
    public function moveImportedFilesToUploads($ret = [], $blogid_to_import = 0, $files_array = [])
    {
        $this->getSystemInitialization()->setSlowProcess();
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Processing upload..', 'prime-mover'));
        
        //Check for diff continuation then skip
        if (isset($ret['diff_confirmation'])) {
            //This is a diff confirmation, skip
            do_action('multisite_migration_log_processed_events', $ret, $blogid_to_import, 'import', 'moveImportedFilesToUploads', $this);
            return $ret;
        }
        //Check if file is set
        if (isset($files_array['file'])) {
            if ( ! empty( $files_array['file'] ) ) {
                $this->getSystemInitialization()->setImportZipPath( $files_array['file'] );
            }
            
            //Append blog_id and import package for easy tracking what site is being imported
            $ret['blog_id']	= $blogid_to_import;
            $this->getSystemInitialization()->setImportBlogID($blogid_to_import);
            $ret['file'] = $files_array['file'];
            
            //Return
            do_action('multisite_migration_log_processed_events', $ret, $blogid_to_import, 'import', 'moveImportedFilesToUploads', $this);
            return $ret;
        }
    }
    /**
     * Unzip upload file
     * {@inheritDoc}
     * @see PrimeMoverImport::unzipImportedZipPackageMigration()
     * @compatible 5.6
     */
    public function unzipImportedZipPackageMigration($ret = [], $blogid_to_import = 0, $files_array = [])
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $this->getSystemInitialization()->setSlowProcess();
        $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Unzipping package', 'prime-mover'));
        
        //Check for diff continuation then skip
        if (isset($ret['diff_confirmation'])) {
            //This is a diff confirmation, skip
            return $ret;
        }
        
        if (isset($ret['error'])) {
            //Errors now detected, bail out and inform user
            return $ret;
        }
        //Proceed, files should be moved to uploads directory already at this point
        global $wp_filesystem;
        $unzip_result	= false;
        
        //Check if file exist
        if ((isset($ret['file'])) && ($wp_filesystem->exists($ret['file']))) {
            //Exist, unzip
            $file_path		=  $ret['file'];            
            $destination	= dirname($file_path) . DIRECTORY_SEPARATOR;
            
            //Get correct extracted zip folder name
            $correct_folder_name = $this->getSystemFunctions()->getCorrectFolderNameFromZip($file_path);
            
            if (! $correct_folder_name) {
                //Errors detected when getting folder, bail out early
                $ret['error']	= esc_html__('Unzip failed. The package is not a multisite migration site package.', 'prime-mover');
                return $ret;
            }            
            $unzipped_directory	= $destination . $correct_folder_name . DIRECTORY_SEPARATOR;
            $this->getSystemInitialization()->setTemporaryImportPackagePath($unzipped_directory);

            $unzip_result = $this->getSystemFunctions()->unzipFileChunked($file_path, $destination, $blogid_to_import);

            if ( ! $unzip_result) {
                //Errors detected during unzipped, bail out early
                $ret['error']	= esc_html__('Unzip failed. Please check if you have zip extension enabled or enable error logs.', 'prime-mover');
                return $ret;
            }

            //Unzip success            
            if (! $wp_filesystem->exists($unzipped_directory)) {
                $ret['error']	= esc_html__('Unzipped directory does not exist.', 'prime-mover');
                return $ret;
            }
            $ret['unzipped_directory']	= $unzipped_directory;
            $this->getSystemInitialization()->setTemporaryImportPackagePath($unzipped_directory);
        } else {
            $ret['error']	= esc_html__('Import file does not seem to exist.', 'prime-mover');
        }
        
        //It's safe to delete the zip file at this point
        $delete_zip = apply_filters('multisite_migration_delete_zip_after_unzip', true, $file_path, $ret);
        if ($delete_zip) {
            $this->getSystemFunctions()->multisiteMigrationDoDelete($file_path);
        }        
        do_action('multisite_migration_log_processed_events', $ret, $blogid_to_import, 'import', 'unzipImportedZipPackageMigration', $this);
        return $ret;
    }
    
    /**
     * Validate imported site vs the package to make sure there is no import mismatch.
     * {@inheritDoc}
     * @see PrimeMoverImport::validateImportedSiteVsPackage()
     * @compatible 5.6
     */
    public function validateImportedSiteVsPackage($ret = [], $blogid_to_import = 0, $files_array = [])
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }        
        $this->getSystemInitialization()->setSlowProcess();
        $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Checking import package', 'prime-mover'));
        
        //Check for diff continuation then skip
        if (isset($ret['diff_confirmation'])) {
            //This is a diff confirmation, skip
            return $ret;
        }
        
        if (isset($ret['error'])) {
            //Errors now detected, bail out and inform user
            return $ret;
        }
        
        //Step1, retrieved the unzip directory name and extract the imported site from this
        if (isset($ret['unzipped_directory'])) {
            $unzipped_directory	= $ret['unzipped_directory'];
            $directory_name		= basename($unzipped_directory);
            $exploded			= explode("_", $directory_name);
            $reversed			= array_reverse($exploded);
            if (isset($reversed[0])) {
                $associated_site_id_of_package	= intval($reversed[0]);
                if ($associated_site_id_of_package > 0) {
                    $blogid_to_import	= intval($blogid_to_import);
                    if ($blogid_to_import !==  $associated_site_id_of_package) {
                        //Mismatch
                        $ret['error']	= esc_html__('Mismatch import!', 'prime-mover');
                    }
                } else {
                    //Corrupted package maybe
                    $ret['error']	= esc_html__('Corrupt package', 'prime-mover');
                }
            } else {
                //Corrupted package maybe
                $ret['error']	= esc_html__('Corrupt package', 'prime-mover');
            }
        } else {
            $ret['error']	= esc_html__('Unzipped directory does not exist.', 'prime-mover');
        }
        
        if (isset($ret['error'])) {
            //Error occurs, delete the temporary import package.
            $this->getSystemFunctions()->multisiteMigrationDoDelete($unzipped_directory);
        }
        
        //API for further validation of the imported package        
        $ret = apply_filters('multisite_migration_validate_imported_package', $ret, $blogid_to_import);
   
        do_action('multisite_migration_log_processed_events', $ret, $blogid_to_import, 'import', 'validateImportedSiteVsPackage', $this);
        
        //Successful import should not contain errors
        return $ret;
    }
    
    /**
     * Compare system footprint
     * {@inheritDoc}
     * @see PrimeMoverImport::compareSystemFootprintImport()
     * @compatible 5.6
     */
    public function compareSystemFootprintImport($ret = [], $blogid_to_import = 0, $files_array = [])
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $this->getSystemInitialization()->setSlowProcess();
        $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Comparing system footprint', 'prime-mover'));
        
        //Check for diff continuation then skip
        if (isset($ret['diff_confirmation'])) {
            //This is a diff confirmation, skip
            return $ret;
        }
        
        if (isset($ret['error'])) {
            //Errors now detected, bail out and inform user
            return $ret;
        }
        
        global $wp_filesystem;
        //Success extraction of folder and site matched must already be checked at this point
        /*
         * Read footprint of imported package
         */
        if (isset($ret['unzipped_directory'])) {
            $unzipped_directory	= $ret['unzipped_directory'];
            $footprint_package	= $unzipped_directory . 'footprint.json';
            if ($wp_filesystem->exists($footprint_package)) {
                //Footprint package exist
                $json_string	=	file_get_contents($footprint_package);
                
                //Convert to array
                $system_footprint_package_array	= json_decode($json_string, true);
                
                //Store to $ret array
                $ret['imported_package_footprint'] = $system_footprint_package_array;
                
                //Generate a system footprint of the target site
                $system_footprint_target_site	= $this->getSystemChecks()->multisiteMigrationGenerateSystemFootprint($blogid_to_import);
                
                //Store information about origin and target site URLs and paths to $ret variable
                if (!isset($system_footprint_package_array['site_url'])) {
                    $ret['error']			= esc_html__('Site URL is not defined in this imported package, please re-generate the package at the source site.', 'prime-mover');
                    return $ret;
                }
                if (!isset($system_footprint_package_array['wp_root'])) {
                    $ret['error']			= esc_html__('WordPress root is not defined in this imported package, please re-generate the package at the source site.', 'prime-mover');
                    return $ret;
                }
                if (!isset($system_footprint_target_site['site_url'])) {
                    $ret['error']			= esc_html__('Site URL is not defined in this site. Please check your multisite settings.', 'prime-mover');
                    return $ret;
                }
                if (!isset($system_footprint_target_site['wp_root'])) {
                    $ret['error']			= esc_html__('WordPress root is not defined in this site. Please check your multisite settings.', 'prime-mover');
                    return $ret;
                }
                
                //All clear
                $ret['origin_site_url']			= $system_footprint_package_array['site_url'];
                $ret['origin_wp_root']			= $system_footprint_package_array['wp_root'];
                $ret['origin_db_prefix']		= $system_footprint_package_array['db_prefix'];
                
                $ret['target_site_url']			= $system_footprint_target_site['site_url'];
                $ret['target_wp_root']			= $system_footprint_target_site['wp_root'];
                $ret['target_db_prefix']		= $system_footprint_target_site['db_prefix'];
                
                //Can be filtered prior to array comparison
                $system_footprint_package_array	= apply_filters('multisite_migration_filter_pre_comparison_package_array', $system_footprint_package_array, $ret, $blogid_to_import);
                $system_footprint_target_site	= apply_filters('multisite_migration_filter_pre_comparison_target_array', $system_footprint_target_site, $ret, $blogid_to_import);

                if ($system_footprint_target_site != $system_footprint_package_array) {
                    //Unequal system footprint
                    $diff			= $this->getSystemFunctions()->getDetailedDifferencesBetweenFootprints($system_footprint_package_array, $system_footprint_target_site);
                    $ret['diff']	= $diff;
                    $ret = apply_filters('multisite_migration_filter_config_after_diff_check', $ret, $blogid_to_import);
                }
            } else {
                //Footprint does not exist
                $ret['error']			= esc_html__('System footprint of package is not found', 'prime-mover');
            }
        }
        do_action('multisite_migration_log_processed_events', $ret, $blogid_to_import, 'import', 'compareSystemFootprintImport', $this);
        return $ret;
    }
    
    /**
     * Remove site URL and site paths before comparison
     * @param array $system_footprint_package_array
     * @param array $ret
     * @param number $blogid_to_import
     * @return array
     * @compatible 5.6
     */
    public function filterPreComparisonPackageArrayFunc($system_footprint_package_array = [], $ret = [], $blogid_to_import = 0)
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        if (! empty($system_footprint_package_array['site_url'])
                && ! empty($system_footprint_package_array['wp_root']) &&
                ! empty($system_footprint_package_array['db_prefix'])) {
            unset($system_footprint_package_array['site_url']);
            unset($system_footprint_package_array['wp_root']);
            unset($system_footprint_package_array['db_prefix']);
        }
        return $system_footprint_package_array;
    }
    
    /**
     * Remove site URL and site paths before comparison
     * @param array $system_footprint_target_site
     * @param array $ret
     * @param number $blogid_to_import
     * @return array
     * @compatible 5.6
     */
    public function filterPreComparisonTargetArrayFunc($system_footprint_target_site = [], $ret = [], $blogid_to_import = 0)
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        if (! empty($system_footprint_target_site['site_url'])
                && ! empty($system_footprint_target_site['wp_root'])
                && ! empty($system_footprint_target_site['db_prefix'])) {
            unset($system_footprint_target_site['site_url']);
            unset($system_footprint_target_site['wp_root']);
            unset($system_footprint_target_site['db_prefix']);
        }
        return $system_footprint_target_site;
    }
    
    /**
     * Update blogs.dir media files with the newly ones being imported
     * {@inheritDoc}
     * @see PrimeMoverImport::updateTargetMediaFilesWithNew()
     * @compatible 5.6
     */
    public function updateTargetMediaFilesWithNew($ret = [], $blogid_to_import = 0, $files_array = [])
    {
        $this->getSystemInitialization()->setSlowProcess();
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        global $wp_filesystem;
        $unzipped_directory	= '';        
        if (isset($ret['error'])) {
            //Errors now detected, bail out and inform user
            return $ret;
        } elseif (isset($ret['diff'])) {
            /**
             * After diff checks, check for 'diff'
             *
             */
            //Diffs exist from previous processing, bail out and inform user
            $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Difference detected, alert user.', 'prime-mover'));
            return $ret;
        } else {
            /**
             * There is no diff and no error, check for unzipped directory
             */
            if ($ret['unzipped_directory']) {
                $unzipped_directory	= $ret['unzipped_directory'];
            }
        }
        $this->getSystemFunctions()->temporarilyIncreaseMemoryLimits();
        $media_path = $unzipped_directory . 'media.zip';        
        
        //Verify if we are importing media files
        if (false === $wp_filesystem->exists($media_path)) {
            //No media files to import, SKIP..
            $ret['skipped_media'] = true;
            return $ret;
        }   

        $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Importing media files', 'prime-mover'));
        $zip = new \ZipArchive();
        $zip_status = $zip->open($media_path);  
        $extraction_failed = esc_html__('Extraction failed.', 'prime-mover');
        if (true === $zip_status) {           
            $zip = $this->getSystemInitialization()->setEncryptionPassword($zip);
            if ( ! $zip->extractTo($unzipped_directory)) {
                $error = $zip->getStatusString();
                if ($error) {
                    $extraction_failed = $extraction_failed . ' ' . $error;
                }                
                $ret['error'] = $extraction_failed;
                return $ret;
            }
            $zip->close();
        } else {
            $ret['error'] = esc_html__('Failed opening archive', 'prime-mover');
            return $ret;
        }        
        
        //Applies to both continuing import mode and validated mode.
        //Delete old directory
        $processed_blog	= 0;
        if ($blogid_to_import > 0) {
            $processed_blog	= $blogid_to_import;
        } elseif (isset($data_to_continue['blog_id'])) {
            $processed_blog	= $data_to_continue['blog_id'];
        }
        
        if ($processed_blog < 0) {
            $ret['error']	= esc_html__('Blog id does not exist.', 'prime-mover');
            return $ret;
        }
        
        //Formulate source directory
        $source	= $unzipped_directory . $processed_blog . DIRECTORY_SEPARATOR;

        /**
         * Check if we can now process this import step
         */
        if (empty($unzipped_directory)) {
            $ret['error']	= esc_html__('Unzipped directory is empty.', 'prime-mover');
            return $ret;
        }
        if (false === $wp_filesystem->exists($unzipped_directory)) {
            //Empty unzipped director and does not exist
            $ret['error']	= esc_html__('Unzipped directory does not exist.', 'prime-mover');
            return $ret;
        }
        
        //Get blogs dir path to delete
        $media_path	= $this->getSystemFunctions()->multisiteMigrationGetBlogsDirPath($processed_blog);
        
        if (false === $wp_filesystem->exists($media_path)) {
            $ret['error']	= esc_html__('Media path directory does not exist.', 'prime-mover');
            return $ret;
        }
        
        //Officially starts the import
        do_action('multisite_migration_before_actual_import', $ret, $blogid_to_import, $files_array);
        
        //Media hash check, skip this process if target and source media files are exactly the same
        $source_hash = $this->getSystemFunctions()->hashEntity($source);
        $target_hash = $this->getSystemFunctions()->hashEntity($media_path);
        if ($source_hash === $target_hash) {
            do_action('multisite_migration_log_processed_events', "MEDIA HASHES MATCHED, NO NEED TO COPY: $source TO: $media_path", $processed_blog, 'import', 'updateTargetMediaFilesWithNew', $this);
            return $ret;           
        }
        if ( is_multisite() ) {
            //This only applies to multisite
            $del_result	= $wp_filesystem->delete($media_path, true);
            if (false === $del_result) {
                $ret['error']	= esc_html__('Error deleting media directory. Please check for permissions.', 'prime-mover');
                return $ret;
            }
        }
        //Media directory of this sub-site deleted, proceed to copy with new directory
        $target	= '';
        if (wp_mkdir_p($media_path)) {
            $target	= $media_path;
        }
        $exportdir_slug = $this->getSystemInitialization()->getMultisiteExportFolderSlug();
        $skip = [$exportdir_slug];
        if (is_multisite()) {
            $skip = [];
        }
        if (($wp_filesystem->exists($source)) && ($wp_filesystem->exists($target))) {
            do_action('multisite_migration_log_processed_events', "MIGRATING MEDIA FILES: $source TO: $target", $processed_blog, 'import', 'updateTargetMediaFilesWithNew', $this);
            $copy_directory_result	= $this->getSystemChecks()->getSystemCheckUtilities()->copyDir($source, $target, $skip);
            if (is_wp_error($copy_directory_result)) {
                $ret['error']= $copy_directory_result->get_error_message();
                return $ret;
            }
        }
        
        /**
         * In case of re-confirmation
         */
        if (empty($ret)) {
            $ret = $data_to_continue;
        }
        do_action('multisite_migration_log_processed_events', $ret, $blogid_to_import, 'import', 'updateTargetMediaFilesWithNew', $this);
        return $ret;
    }
    
    /**
     * {@inheritDoc}
     * @see PrimeMoverImport::importDb()
     * @compatible 5.6
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverImporter::itImportsUnencryptedDbWhenRequested()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverImporter::itImportsEncryptedDbIfAllSet()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverImporter::itDoesNotImportWhenNotAuthorized()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverImporter::itDoesNotImportWhenItHasDiffs()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverImporter::itDoesNotImportDbWhenItHasErrors()
     */
    public function importDb($ret = [], $blogid_to_import = 0, $files_array = [])
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $this->getSystemInitialization()->setSlowProcess();
        if ((isset($ret['error'])) || (isset($ret['diff']))) {
            //Errors or diff still detected, bail out and inform user
            $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Errors or diffs detected, alert user.', 'prime-mover'));
            return $ret;
        }
        if (!isset($ret['unzipped_directory'])) {
            $ret['error']	= esc_html__('Unzipped directory is missing. It is deleted while the import is ongoing. Please repeat again.', 'prime-mover');
            return $ret;
        }
        //Proceed.
        $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Importing database tables', 'prime-mover'));

        $unzipped_path = $ret['unzipped_directory'];
        $import_path_sql = $this->handleEncrypteddB($unzipped_path, $blogid_to_import);        
        if ( ! $import_path_sql ) {
            $ret['error']		= esc_html__('Unable to find the correct SQL path.', 'prime-mover');
            return $ret;
        }

        $import_command	= $this->mysqlRestoreCommand($import_path_sql);
        
        do_action('multisite_migration_log_processed_events', 'MySQL import command: ', $blogid_to_import, 'import', 'importDb', $this);
        do_action('multisite_migration_log_processed_events', $import_command, $blogid_to_import, 'import', 'importDb', $this);
        do_action('multisite_migration_before_db_processing', $ret, $blogid_to_import);
        
        $import_result = $this->restoredB($import_command);
        if ($import_result) {
            $ret['error'] = $import_result;
            return $ret;
        }
        do_action('multisite_migration_log_processed_events', $ret, $blogid_to_import, 'import', 'importDb', $this);
        return $ret;
    }

    /**
     * Get correct import SQL depending whether a raw SQL is available or encrypted format
     * @param string $unzipped_path
     * @param number $blogid_to_import
     * @return string
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverImporter::itImportsUnencryptedDbWhenRequested()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverImporter::itImportsEncryptedDbIfAllSet()
     */
    private function handleEncrypteddB($unzipped_path = '', $blogid_to_import = 0)
    {
        global $wp_filesystem;
        $import_path_sql = '';
        if ( ! $unzipped_path || ! $blogid_to_import ) {
            return $import_path_sql;
        }
        $db_import_options = ['.sql', '.sql.enc'];
        foreach ($db_import_options as $extension ) {
            $sql_file_name = $blogid_to_import . $extension;
            $temp_path_sql = $unzipped_path . $sql_file_name;            
            
            if ($wp_filesystem->exists($temp_path_sql)) {
                $import_path_sql = $temp_path_sql;
            }            
            if ($import_path_sql && '.sql.enc' === $extension) {
                $this->getSystemInitialization()->setEncryptedDb(true);
            }
            if ($import_path_sql) {
                return $import_path_sql;
            }
        }
        return $import_path_sql;
    }

    /**
     * Formulate MySQL restore command with encryption support.
     * @param string $sql_file_name
     * @return string[]|boolean[]
     */
    protected function mysqlRestoreCommand($sql_file_name = '')
    {   
        $commands = [];
        $decrypt = false;
        if ($this->getSystemInitialization()->getEncryptedDb()) {
            $decrypt = true;
        }      
        
        $commands['target_path'] = $sql_file_name;
        $commands['decrypt'] = $decrypt;
        
        return $commands;        
    }
    
    /**
     * Execute database restore
     * @param array $restore_commands
     * @return boolean|string
     */
    protected function restoredB($restore_commands = []) 
    {
        if ( empty($restore_commands) || empty($restore_commands['target_path']) || ! isset($restore_commands['decrypt']) ) {
            return false;
        }
        $target_path = $restore_commands['target_path'];
        $decrypt = $restore_commands['decrypt'];

        $handle = @fopen($target_path, 'r');
        if ( ! $handle) {
            return esc_html__('Unable to open database file.', 'prime-mover');
        }         
        global $wpdb;
        $templine = '';       
        while(!feof($handle)){
            $line = fgets($handle);
            if ($decrypt) {
                $line = apply_filters('multisite_migration_decrypt_data', $line);
            }            
            
            if (substr($line, 0, 2) == '--' || $line == '') {
                continue;
            }
            
            $templine .= $line;            
            if (substr(trim($line), -1, 1) == ';') {                
                $wpdb->query($templine);                    
                $templine = '';
            }            
        }   

        fclose($handle);        
    }
    
    /**
     *
     * @param array $converted_table_names
     * @compatible 5.6
     */
    private function renameDbPrefixHelper($converted_table_names = [])
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        if (is_array($converted_table_names) && ! empty($converted_table_names)) {
            global $wpdb;
            foreach ($converted_table_names as $original_table_name => $new_table_name) {
                
                $drop_query = sprintf(
                    "DROP TABLE IF EXISTS `%s`;",
                    $new_table_name
                    );
                
                $wpdb->query($drop_query);                
                
                $rename_query = sprintf(
                        "RENAME TABLE `%s` TO `%s`;",
                        $original_table_name,
                        $new_table_name
                        );
                
                $wpdb->query($rename_query);
            }
        }
    }
    
    /**
     * Rename table names to have a compatible dB prefix if possible
     * {@inheritDoc}
     * @see PrimeMoverImport::renameDbPrefix()
     * @compatible 5.6
     */
    public function renameDbPrefix($ret = [], $blogid_to_import = 0, $files_array = [])
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $this->getSystemInitialization()->setSlowProcess();
        if ((isset($ret['error'])) || (isset($ret['diff']))) {            
            $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Errors or diffs detected, alert user.', 'prime-mover'));
            return $ret;
        }
        if ($ret['origin_db_prefix'] === $ret['target_db_prefix']) {            
            return $ret;
        }
        $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Renaming dB prefix with target site.', 'prime-mover'));
        
        global $wpdb;
        $db_search = "SHOW TABLES LIKE %s";        
        
        $clean_tables = [];  
        $package_exported_tables = $ret['imported_package_footprint']['exported_db_tables'];
        if ( ! is_array($package_exported_tables) ) {
            return $ret;
        }
        foreach ($package_exported_tables as $table_to_validate) {
            $table_to_validate = filter_var($table_to_validate, FILTER_SANITIZE_STRING);
            $sql = $wpdb->prepare($db_search ,  $table_to_validate);
            if ($wpdb->get_var($sql) === $table_to_validate) {
                $clean_tables[] = $table_to_validate;
            }
        }
    
        if (empty($clean_tables)) {
            $ret['error']		= esc_html__('Unable to rename tables, please check that your database is not empty or these tables exists.', 'prime-mover');
            return $ret;
        }

        $converted_table_names = $this->convertOldTableNamesToNew($ret, $clean_tables);  
        $this->renameDbPrefixHelper($converted_table_names);
        $this->renameSomeOptions($ret, $blogid_to_import);
 
        do_action('multisite_migration_log_processed_events', $ret, $blogid_to_import, 'import', 'renameDbPrefix', $this);
        do_action('multisite_migration_after_renamedb_prefix', $blogid_to_import, $ret);
        
        return $ret;
    }
    
    /**
     * Activate plugins if not still activated
     * {@inheritDoc}
     * @see \Codexonics\PrimeMoverFramework\interfaces\PrimeMoverImport::activatePluginsIfNotActivated()
     * @compatible 5.6
     */
    public function activatePluginsIfNotActivated($ret = [], $blogid_to_import = 0, $files_array = [])
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $this->getSystemInitialization()->setSlowProcess();
        if ((isset($ret['error'])) || (isset($ret['diff']))) {
            //Errors or diff still detected, bail out and inform user
            $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Errors or diffs detected, alert user.', 'prime-mover'));
            return $ret;
        }
        
        $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Activating plugins', 'prime-mover'));
        if (! isset($ret['imported_package_footprint']['plugins'])) {
            return $ret;
        }

        //Get activated plugins in the source site
        $source_site_active_plugins = $ret['imported_package_footprint']['plugins'];
        if (empty($source_site_active_plugins) || ! is_array($source_site_active_plugins)) {
            return $ret;
        }
        //Switch to current blog
        $this->getSystemFunctions()->switchToBlog($blogid_to_import);
        wp_cache_delete('alloptions', 'options');
        global $wp_filesystem;
        foreach ($source_site_active_plugins as $plugin => $version) {
            if (empty($plugin)) {
                continue;
            }
            if (! $wp_filesystem->exists(PRIME_MOVER_PLUGIN_CORE_PATH . $plugin)) {
                continue;
            }
            if (is_plugin_inactive($plugin)) {
                activate_plugin($plugin, '', false, true);
            }
        }
        
        //Restore current blog
        $this->getSystemFunctions()->restoreCurrentBlog();
        
        do_action('multisite_migration_after_db_processing', $ret, $blogid_to_import);
        do_action('multisite_migration_log_processed_events', $ret, $blogid_to_import, 'import', 'activatePluginsIfNotActivated', $this);
        
        return $ret;
    }

    /**
     * Success import
     * {@inheritDoc}
     * @see \Codexonics\PrimeMoverFramework\interfaces\PrimeMoverImport::markImportSuccess()
     * @compatible 5.6
     */
    public function markImportSuccess($ret = [], $blogid_to_import = 0, $files_array = [])
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $this->getSystemInitialization()->setSlowProcess();
        if ((isset($ret['error'])) || (isset($ret['diff']))) {
            //Errors or diff still detected, bail out and inform user
            $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Errors or diffs detected, alert user.', 'prime-mover'));
            return $ret;
        }
        
        //Officially ends the import
        do_action('multisite_migration_after_actual_import', $ret, $blogid_to_import, $files_array);
        do_action('multisite_migration_log_processed_events', 'Marked import success', $blogid_to_import, 'import', 'markImportSuccess', $this);
        
        //Import success
        $ret['import_success']	= true;
        do_action('multisite_migration_log_processed_events', $ret, $blogid_to_import, 'import', 'markImportSuccess', $this);
        
        return $ret;
    }
    
    /**
     * Rename some options in target dB after renaming dB prefix
     * @param array $ret
     * @param number $blogid_to_import
     * @return void
     * @compatible 5.6
     */
    private function renameSomeOptions($ret = [], $blogid_to_import = 0)
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        if (empty($ret) || ! $blogid_to_import) {
            return $ret;
        }
        
        $affected_options = $this->getAffectedOptionsAfterRenamePrefix($ret, $blogid_to_import);
        if (empty($affected_options)) {
            return $ret;
        }
        $this->executeRenameOptionsQuery($affected_options, $blogid_to_import, $ret);
    }
    
    /**
     * Remove prefix from table name
     * @param string $prefix
     * @param string $table
     * @return string
     */
    private function removePrefixFromTable($prefix = '', $table = '')
    {        
        if (substr($table, 0, strlen($prefix)) == $prefix) {
            $table = substr($table, strlen($prefix));
        }        
        return $table;
    }
    
    /**
     * Execute rename options query
     * @param array $affected_options
     * @param number $blogid_to_import
     * @param array $ret
     * @compatible 5.6
     */
    private function executeRenameOptionsQuery($affected_options = [], $blogid_to_import = 0, $ret = [])
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        if (empty($affected_options) || ! is_array($affected_options) || ! $blogid_to_import) {
            return;
        }
        global $wpdb;
        
        //Switch to this blog
        $this->getSystemFunctions()->switchToBlog($blogid_to_import);
        
        $origin_prefix = $ret['origin_db_prefix'];        
        foreach ($affected_options as $v) {            
            $name_without_prefix = $this->removePrefixFromTable($origin_prefix, $v);
            $new_option_name = $wpdb->prefix . $name_without_prefix;
            delete_option($new_option_name);
            
            $update_query = "UPDATE {$wpdb->prefix}options SET option_name = '{$new_option_name}' WHERE option_name = '%s'";
            $wpdb->query($wpdb->prepare($update_query, $v));
        }
        
        //restore current blog
        $this->getSystemFunctions()->restoreCurrentBlog();
    }
    
    /**
     *
     * @param array $ret
     * @param number $blogid_to_import
     * @return void|array|mixed[]
     * @compatible 5.6
     */
    private function getAffectedOptionsAfterRenamePrefix($ret = [], $blogid_to_import = 0)
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $affected_options = array();
        if (empty($ret) || ! $blogid_to_import) {
            return $affected_options;
        }
        
        //Switch to this blog
        $this->getSystemFunctions()->switchToBlog($blogid_to_import);
        
        global $wpdb;
        
        $options_query = "SELECT option_name FROM {$wpdb->prefix}options WHERE option_name LIKE %s";
        $prefix_search	= $wpdb->esc_like($ret['origin_db_prefix']) . '%';
        $option_query_prepared = $wpdb->prepare($options_query, $prefix_search);
        $option_query_results = $wpdb->get_results($option_query_prepared, ARRAY_N);
        
        if (! is_array($option_query_results) || empty($option_query_results)) {
            return $affected_options;
        }
        foreach ($option_query_results as $v) {
            if (! is_array($v)) {
                continue;
            }
            $affected_options[] = reset($v);
        }
        
        //restore current blog
        $this->getSystemFunctions()->restoreCurrentBlog();
        
        return $affected_options;
    }
    /**
     * Convert old table names to new
     * @param array $ret
     * @param array $clean_tables
     * @return mixed[]
     * @compatible 5.6
     */
    private function convertOldTableNamesToNew($ret = [], $clean_tables = [])
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $converted = array();
        if (! empty($clean_tables)) {
            //Loop over clean table names
            foreach ($clean_tables as $k => $original_table_name) {
                $new_table_name = str_replace($ret['origin_db_prefix'], $ret['target_db_prefix'], $original_table_name);
                $converted[ $original_table_name ] = $new_table_name;
            }
        }
        return $converted;
    }
    
    /**
     * {@inheritDoc}
     * @see PrimeMoverImport::searchAndReplace()
     * @compatible 5.6
     */
    public function searchAndReplace($ret = [], $blogid_to_import = 0, $files_array = [])
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $this->getSystemInitialization()->setSlowProcess();
        if ((isset($ret['error'])) || (isset($ret['diff']))) {
            //Errors or diff still detected, bail out and inform user
            $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Errors or diffs detected, alert user.', 'prime-mover'));
            return $ret;
        }       
        
        //Generate replaceables
        $replaceables = $this->getSystemFunctions()->generateImportReplaceables($ret);
        
        if (apply_filters('multisite_migration_skip_search_replace', false, $ret, $blogid_to_import)) {
            do_action('multisite_migration_skipped_search_replace', $ret, $blogid_to_import, $replaceables);
            return $ret;
        }
        
        $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Search and replace.', 'prime-mover'));
        
        do_action('multisite_migration_log_processed_events', "Starting search and replace with these following parameters", $blogid_to_import, 'import', 'searchAndReplace', $this);
        do_action('multisite_migration_log_processed_events', $replaceables, $blogid_to_import, 'import', 'searchAndReplace', $this);
        do_action('multisite_migration_log_processed_events', "Footprint parameters: ", $blogid_to_import, 'import', 'searchAndReplace', $this);
        do_action('multisite_migration_log_processed_events', $ret, $blogid_to_import, 'import', 'searchAndReplace', $this);
        
        if ((is_array($replaceables)) && (empty($replaceables))) {
            $ret['error']		= esc_html__('Unable to generate replaceables, could that your site package is corrupted. Please check.', 'prime-mover');
            return $ret;
        }
        
        //Switch to current blog
        $this->getSystemFunctions()->switchToBlog($blogid_to_import);
        @set_time_limit(0);
        @ini_set('memory_limit', '2048M');
        
        $connection = $this->getSystemInitialization()->getConnectionInstance();
        $all_tables = $this->getSystemInitialization()->getTablesforReplacement();
        
        global $wpdb;
        $posts_table = $wpdb->posts;
        
        $excluded_columns = [$posts_table => 'guid']; 
        PrimeMoverSearchReplace::load($connection, $replaceables, $all_tables, 1, $this, $excluded_columns);
 
        //Restore current blog
        $this->getSystemFunctions()->restoreCurrentBlog();        
        
        return $ret;
    }
    
    /**
     * {@inheritDoc}
     * @see \Codexonics\PrimeMoverFramework\interfaces\PrimeMoverImport::markTargetSiteUploadsInformation()
     * @compatible 5.6
     */
    public function markTargetSiteUploadsInformation($ret = [], $blogid_to_import = 0, $files_array = [])
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $this->getSystemInitialization()->setSlowProcess();
        if ((isset($ret['error'])) || (isset($ret['diff']))) {
            //Errors or diff still detected, bail out and inform user
            $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Errors or diffs detected, alert user.', 'prime-mover'));
            return $ret;
        }
        
        $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Saving uploads info..', 'prime-mover'));
        $upload_path_information = '';
        $upload_url_path_information = '';
        
        //Get 'uploads_path' info in dB
        $uploads_path = $this->getSystemFunctions()->getBlogOption($blogid_to_import, 'upload_path');
        
        //Get 'upload_url_path' info in dB
        $upload_url_path = $this->getSystemFunctions()->getBlogOption($blogid_to_import, 'upload_url_path');
        
        if ($uploads_path) {
            //Set, might be custom
            $upload_path_information = $uploads_path;
        }
        
        if ($upload_url_path) {
            $upload_url_path_information = $upload_url_path;
        }

        $ret['upload_path_information'] = $upload_path_information;
        $ret['upload_url_path_information'] = $upload_url_path_information;
        
        $ret = apply_filters('multisite_migration_filter_other_information', $ret, $blogid_to_import);
        
        do_action('multisite_migration_log_processed_events', $ret, $blogid_to_import, 'import', 'markTargetSiteUploadsInformation', $this);
        return $ret;
    }

    /**
     * {@inheritDoc}
     * @see \Codexonics\PrimeMoverFramework\interfaces\PrimeMoverImport::restoreCurrentUploadsInformation()
     * @compatible 5.6
     */
    public function restoreCurrentUploadsInformation($ret = [], $blogid_to_import = 0, $files_array = [])
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $this->getSystemInitialization()->setSlowProcess();
        if ((isset($ret['error'])) || (isset($ret['diff']))) {
            //Errors or diff still detected, bail out and inform user
            $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Errors or diffs detected, alert user.', 'prime-mover'));
            return $ret;
        }
        
        $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Restoring uploads info..', 'prime-mover'));
        
        if (! isset($ret['upload_path_information']) || ! isset($ret['upload_url_path_information'])) {
            return $ret;
        }

        $this->getSystemFunctions()->switchToBlog($blogid_to_import);
        wp_cache_delete('alloptions', 'options');
        $this->getSystemFunctions()->restoreCurrentBlog();
        
        $upload_structure = $ret['upload_path_information'];
        $upload_url = $ret['upload_url_path_information'];
        
        $this->getSystemFunctions()->updateBlogOption($blogid_to_import, 'upload_path', $upload_structure);
        $this->getSystemFunctions()->updateBlogOption($blogid_to_import, 'upload_url_path', $upload_url);
 
        do_action('multisite_migration_log_processed_events', $ret, $blogid_to_import, 'import', 'restoreCurrentUploadsInformation', $this);
        
        return $ret;
    }
}
