<?php
namespace Codexonics\PrimeMoverFramework\classes;

/*
 * This file is part of the Codexonics.PrimeMoverFramework package.
 *
 * (c) Codexonics Ltd
 *
 * This package is Open Source Software. For the full copyright and license
 * information, please view the LICENSE file which was distributed with this
 * source code.
 */

use Codexonics\PrimeMoverFramework\interfaces\PrimeMoverSystemProcessor;
use Codexonics\PrimeMoverFramework\utilities\PrimeMoverExportUtilities;
use Codexonics\PrimeMoverFramework\utilities\PrimeMoverUploadUtilities;

if (! defined('ABSPATH')) {
    exit;
}

/**
 * Prime Mover System Processors Class
 *
 * The Prime Mover System Processor Class aims to provide the AJAX processor facility for exporting,importing and deleting temporary files.
 */
class PrimeMoverSystemProcessors implements PrimeMoverSystemProcessor
{
    private $system_authorization;
    private $system_checks;
    private $system_initialization;
    private $system_functions;
    private $importer;
    private $upload_utilities;
    private $export_utilities;
    private $progress_handlers;
    
    /**
     * Constructor
     * @param PrimeMoverSystemAuthorization $system_authorization
     * @param PrimeMoverSystemChecks $system_checks
     * @param PrimeMoverSystemInitialization $system_initialization
     * @param PrimeMoverSystemFunctions $system_functions
     * @param PrimeMoverImporter $importer
     * @param PrimeMoverUploadUtilities $upload_utilities
     * @param PrimeMoverExportUtilities $export_utilities
     * @param PrimeMoverProgressHandlers $progress_handlers
     */
    public function __construct(
        PrimeMoverSystemAuthorization $system_authorization,
        PrimeMoverSystemChecks $system_checks,
        PrimeMoverSystemInitialization $system_initialization,
        PrimeMoverSystemFunctions $system_functions,
        PrimeMoverImporter $importer,
        PrimeMoverUploadUtilities $upload_utilities,
        PrimeMoverExportUtilities $export_utilities,
        PrimeMoverProgressHandlers $progress_handlers
    ) 
    {
        $this->system_authorization = $system_authorization;
        $this->system_checks = $system_checks;
        $this->system_initialization = $system_initialization;
        $this->system_functions = $system_functions;
        $this->importer = $importer;
        $this->upload_utilities = $upload_utilities;
        $this->export_utilities = $export_utilities;
        $this->progress_handlers = $progress_handlers;
    }
 
    /**
     * Get progress handlers
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMoverProgressHandlers
     * @compatible 5.6
     */
    public function getProgressHandlers() 
    {
        return $this->progress_handlers;
    }
    
    /**
     * Get export utilities
     * @return \Codexonics\PrimeMoverFramework\utilities\PrimeMoverExportUtilities
     * @compatible 5.6
     */
    public function getExportUtilities()
    {
        return $this->export_utilities;
    }
    
    /**
     *
     * Get upload utilities
     * @return \Codexonics\PrimeMoverFramework\utilities\PrimeMoverUploadUtilities
     * @compatible 5.6
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverSystemProcessors::itProcessUploadsWhenAllSet() 
     */
    public function getUploadUtilities()
    {
        return $this->upload_utilities;
    }
    /**
     *
     * Get Importer
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMoverImporter
     * @compatible 5.6
     */
    public function getImporter()
    {
        return $this->importer;
    }
    
    /**
     *
     * Get system functions
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMoverSystemFunctions
     * @compatible 5.6
     */
    public function getSystemFunctions()
    {
        return $this->system_functions;
    }
    
    /**
     *
     * Get System Initialization
     * @compatible 5.6
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMoverSystemInitialization
     */
    public function getSystemInitialization()
    {
        return $this->system_initialization;
    }
    
    /**
     *
     * Get System checks
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMoverSystemChecks
     * @compatible 5.6
     */
    public function getSystemChecks()
    {
        return $this->system_checks;
    }
    
    /**
     *
     * Get System authorization
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMoverSystemAuthorization
     * @compatible 5.6
     */
    public function getSystemAuthorization()
    {
        return $this->system_authorization;
    }

    /**
     *
     * Verify encrypted package as requested from client side
     * @compatible 5.6
     */
    public function primeMoverVerifyEncryptedPackage()
    {
        $response = [];
        if ( ! $this->getSystemAuthorization()->isUserAuthorized()) {
            $response['result'] = false;
            $response['error'] = __('Unauthorized request', 'prime-mover');
            return wp_send_json($response);
        }

        $args = [
            'prime_mover_decryption_check_nonce' => FILTER_SANITIZE_STRING,
            'stringtocheck' => FILTER_SANITIZE_STRING,
            'encrypted_media' => FILTER_SANITIZE_STRING,
            'multisiteblog_id' => FILTER_SANITIZE_NUMBER_INT
        ];

        $verify_result = $this->getSystemInitialization()->getUserInput('post', $args, 'verify_encrypted_package', '', 0, true, true);
        if ( empty($verify_result['prime_mover_decryption_check_nonce'] ) ) {
            $response['result'] = false;
            $response['error'] = __('Incorrect nonce', 'prime-mover');
            return wp_send_json($response);
        }

        if ( ! wp_verify_nonce($verify_result['prime_mover_decryption_check_nonce'], 'prime_mover_decryption_check_nonce') ) {
            $response['result'] = false;
            $response['error'] = __('Incorrect nonce', 'prime-mover');
            return wp_send_json($response);
        }
        
        if ( empty($verify_result['stringtocheck'] ) && ! empty($verify_result['validation_errors']) ) {
            $response['result'] = false;
            $response['error'] = $verify_result['validation_errors']['stringtocheck'];
            return wp_send_json($response);
        }
        
        $phpinfoarray = $this->getProgressHandlers()->getShutDownUtilities()->phpinfo2array();        
        $can_decrypt_media = apply_filters('prime_mover_can_decrypt_media', true, $verify_result, $phpinfoarray);
        if (false === $can_decrypt_media) {
            $response['result'] = false;
            $response['error'] = $this->getSystemInitialization()->returnCommonMediaDecryptionError();
            return wp_send_json($response);
        }
        $response['result'] = true;
        return wp_send_json($response);
    }
    
    /**
     *
     * Delete temporary file when user cancels the import
     * @compatible 5.6
     */
    public function multisiteMigrationTempfileDeleteProcessor()
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        
        //Initialize response
        $response = [];
        $response['tempfile_deletion_status'] = false;
        
        //Initialize args
        $args = [
            'multisite_migration_deletetmpfile_nonce' => FILTER_SANITIZE_STRING,
            'temp_file_to_delete' => FILTER_SANITIZE_STRING,
            'action' => FILTER_SANITIZE_STRING,
            'diff_reject' => FILTER_SANITIZE_STRING
        ];
        
        //Validate
        $delete_tmpfile_post = $this->getSystemInitialization()->getUserInput('post', $args, 'delete_tmp_dir', '', 0, true);
        if (isset($delete_tmpfile_post['multisite_migration_deletetmpfile_nonce']) &&
            isset($delete_tmpfile_post[ 'temp_file_to_delete' ]) && wp_verify_nonce($delete_tmpfile_post['multisite_migration_deletetmpfile_nonce'], 'multisite_migration_deletetmpfile_nonce')) {
            //Get temp directory to delete
            global $wp_filesystem;
            $temp_dir = $delete_tmpfile_post[ 'temp_file_to_delete' ];
            if ($temp_dir && $wp_filesystem->exists($temp_dir)) {
                $this->getSystemFunctions()->multisiteMigrationDoDelete($temp_dir);
                $response['tempfile_deletion_status'] = true;
            }
        }
        //Other actions
        do_action('multisite_migration_do_things_tmp_deleted', $delete_tmpfile_post);
        
        //Update client of the status        
        wp_send_json($response);
    }    
    
    /**
     *
     * Export progress shutdown
     * @compatible 5.6
     */
    public function multisiteMigrationShutdownExportProcessor()
    {
        $this->getProgressHandlers()->commonShutDownProcessor('export');
    }
    
    /**
     *
     * Import progress shutdown
     * @compatible 5.6
     */
    public function multisiteMigrationShutdownImportProcessor()
    {
        $this->getProgressHandlers()->commonShutDownProcessor('import');
    }
    
    /**
     *
     * Import progress process
     * @compatible 5.6
     */
    public function multisiteMigrationImportProgressProcessor()
    {
        $this->getProgressHandlers()->commonProgressProcessor('import');
    }
 
    /**
     *
     * Export progress process
     * @compatible 5.6
     */
    public function multisiteMigrationExportProgressProcessor()
    {
        $this->getProgressHandlers()->commonProgressProcessor('export');
    }
    
    /**
     * AJAX Processor for "chunk" uploading
     * @compatible 5.6
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverSystemProcessors::itProcessUploadsWhenAllSet() 
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverSystemProcessors::itDoesNotProcessUploadsIfNotAuthorized() 
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverSystemProcessors::itDoesNotProcessUploadsWhenNonceFailed()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverSystemProcessors::itDoesNotProcessUploadsIfNotReallyUploaded()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverSystemProcessors::itDoesNotProcessUploadsIfNotCorrectMimeType()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverSystemProcessors::itProcessMissingChunks() 
     */
    public function multisiteMigrationUploadsProcessor()
    {
        $all_clear	= false;
        
        //Initialize response
        $response = [];
        
        if ( ! $this->getSystemAuthorization()->isUserAuthorized()) {
            $response['status']= false ;
            $response['error'] = esc_js(__('Unauthorized request.', 'prime-mover'));
            
            return wp_send_json($response); 
        }
        
        $uploads_ajax_input = $this->getUploadUtilities()->getUploadsAjaxInput();
        if ( ! empty($uploads_ajax_input['multisite_blogid_to_import']) ) {
            $this->getSystemInitialization()->setImportBlogID($uploads_ajax_input['multisite_blogid_to_import']);
        }
        $all_clear = $this->getUploadUtilities()->validateChunkUploadCall($uploads_ajax_input, $uploads_ajax_input['multisite_blogid_to_import']);
        $isFixingMissingChunk = $this->getUploadUtilities()->isFixingMissingChunk($uploads_ajax_input);
        
        if (false === $all_clear) {
            do_action('multisite_migration_log_processed_events', "Upload chunk validation failed.", $uploads_ajax_input['multisite_blogid_to_import'], 'import', 'multisiteMigrationUploadsProcessor', $this);
            return wp_send_json($response);
        }
        
        /**
         * All clear here
         */
        $files_array = $all_clear;
        
        if (! isset($files_array['file'])) {
            $response['status']= false ;
            $response['error']	= esc_html__('No file uploaded', 'prime-mover');
            
            return wp_send_json($response);
        }
        
        $uploadedfile 		= $files_array['file'];
        $upload_overrides 	= [ 'test_form' => false, 'test_type' => false ];
        $fileName = sanitize_file_name($files_array["file"]["name"]);
        
        // Get which chunk is being received
        $chunk = (int) $uploads_ajax_input['chunk'];
        $chunks = (int) $uploads_ajax_input['chunks'];
        
        do_action('multisite_migration_log_processed_events', "Server side processing chunk: $chunk of $chunks", $uploads_ajax_input['multisite_blogid_to_import'], 'import', 'multisiteMigrationUploadsProcessor', $this);        
        $this->getUploadUtilities()->setChunk($chunk);
        
        $ret = $this->getUploadUtilities()->moveUploadedChunkToUploads($uploadedfile, $upload_overrides, $fileName);
        
        //Check for uploading errors
        global $wp_filesystem;
        if (! isset($ret['file']) || ! $wp_filesystem->exists($ret['file'])) {
            return wp_send_json($response);
        }
        
        if (isset($ret['error'])) {
            //Error exist, terminate upload and inform client.
            $response['status']= false ;
            $response['error']	= esc_html__('Chunk upload error.', 'prime-mover');
            $this->getSystemFunctions()->multisiteMigrationDoDelete($ret['file']);
            return wp_send_json($response);
        }
        
        $filePath = dirname($ret['file']) . DIRECTORY_SEPARATOR . $fileName;
        $done = $this->getUploadUtilities()->maybeReassembleImportPackageChunks($chunks, $chunk, $filePath, $isFixingMissingChunk);        
        
        if (! is_bool($done) && $done > 0) {
            do_action('multisite_migration_log_processed_events', "Reporting missing chunk back to client browser: $done", $uploads_ajax_input['multisite_blogid_to_import'], 'import', 'multisiteMigrationUploadsProcessor', $this);
            $response['missing_chunk'] = $done;
            return wp_send_json($response);
        }
        
        while($this->getUploadUtilities()->maybePackageIsUnderAssemblyNow($uploads_ajax_input['multisite_blogid_to_import'])) {
            sleep(1);
        }
        
        if (true === $done && ! $this->getUploadUtilities()->isReallyZip($filePath)) {
            $response['status']= false ;
            $response['error'] = esc_html__('File type error.', 'prime-mover');
            
            do_action('multisite_migration_log_processed_events', "File type error DETECTED: $filePath", $uploads_ajax_input['multisite_blogid_to_import'], 'import', 'multisiteMigrationUploadsProcessor', $this);
            $this->getSystemFunctions()->multisiteMigrationDoDelete($filePath);
            return wp_send_json($response);
        }
        
        $response['status']= true ;
        $response['done'] = $done;
        
        $response['chunk']= $chunk;
        $response['chunks']= $chunks;
        $response['filepath'] = $filePath;
        
        do_action('multisite_migration_log_processed_events', "Returning response $chunk of $chunks of $filePath", $uploads_ajax_input['multisite_blogid_to_import'], 'import', 'multisiteMigrationUploadsProcessor', $this); 
        wp_send_json($response);
    }
    
    /**
     * Process import request via AJAX
     * {@inheritDoc}
     * @see PrimeMoverSystemProcessor::multisiteMigrationImportProcessor()
     * @compatible 5.6
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverSystemProcessors::itRunsPrimeMoverImportProcessorIfAllClear()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverSystemProcessors::itDoesNotRunImportProcessorIfNotAuthorized()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverSystemProcessors::itReturnsErrorIfImportProcessorFails()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverSystemProcessors::itReturnsGenericErrorIfImportReturnsUndefined()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverSystemProcessors::itDoesNotRunImportProcessorIfNonceFailed()
     */
    public function multisiteMigrationImportProcessor()
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        
        //Initialize response
        $response		= [];
        
        //Initialize $all_clear
        $all_clear		= false;
        
        //Initialize $ret
        $ret			= [];
        
        //Initialize files array
        $files_array	= [];
        /**
         * Check for Continue handler first
         *
         */
        /**
        * Define sanitizing arguments
        * @var array $args
        */
        
        $args		= [
                'nonce_to_continue'	=> FILTER_SANITIZE_STRING,
                'data_to_continue'	=> FILTER_SANITIZE_STRING,
                'diff_blog_id' => FILTER_SANITIZE_NUMBER_INT
        ];
        
        //Check if we are continuing import after reporting diffs to user.
        $continue_import_post = $this->getSystemInitialization()->getUserInput('post', $args, 'diff_import_processor_ajax', 'import');
        if (isset($continue_import_post['nonce_to_continue'])) {
            //We are continuing import here from a diff confirmation
            if (wp_verify_nonce($continue_import_post['nonce_to_continue'], 'multisite_migration_continue_import')) {
                //Nonce passed
                //Check for in-progress import data being posted
                if ((isset($continue_import_post[ 'data_to_continue' ])) && (!empty($continue_import_post[ 'data_to_continue' ]))) {
                    //Data to continue set
                    $data_to_continue	= $continue_import_post[ 'data_to_continue' ];
                    $data_to_continue	= stripslashes(html_entity_decode($data_to_continue));
                    $data_to_continue	= str_replace('\\', '/', $data_to_continue);
                    $data_to_continue	= json_decode($data_to_continue, true);
                    if ((is_array($data_to_continue)) && (!empty($data_to_continue))) {
                        global $wp_filesystem;
                        //We want to know if the unzipped folder directory exists
                        if ((isset($data_to_continue['unzipped_directory'])) &&
                           (isset($data_to_continue['diff'])) &&
                           (isset($data_to_continue['blog_id']))
                                ) {
                            $unzipped_directory_to_check	= $data_to_continue['unzipped_directory'];
                            //Check if exist
                            if ($wp_filesystem->exists($unzipped_directory_to_check)) {
                                //Remove diff
                                unset($data_to_continue['diff']);
                                //Mark that this process is a continuation
                                $data_to_continue['diff_confirmation']	= true;
                                //Redefine blog ID to import
                                $blogid_to_import	= (int) $data_to_continue['blog_id'];
                                if ($blogid_to_import > 0) {
                                    //We make set the $data_to_continue
                                    $ret		= $data_to_continue;

                                    //And we are clear to go..
                                    do_action('multisite_migration_log_processed_events', 'Diff all cleared and validated, continuing the import', $blogid_to_import, 'import', 'multisiteMigrationImportProcessor', $this);
                                    $all_clear	= true;
                                }
                            }
                        }
                    }
                }
            }
        } else {
            //Original handler
            $args		= [
                    'multisite_migration_import_nonce'		=> FILTER_SANITIZE_STRING,
                    'multisite_blogid_to_import'			=> FILTER_SANITIZE_NUMBER_INT,
                    'multisite_import_package_uploaded_file' => FILTER_SANITIZE_STRING
            ];
            $import_ajax_input = $this->getSystemInitialization()->getUserInput('post', $args, 'import_processor_ajax', 'import');
            if (isset($import_ajax_input['multisite_migration_import_nonce'])) {
                if (wp_verify_nonce($import_ajax_input['multisite_migration_import_nonce'], 'multisite_migration_import_nonce')) {
                    //Nonce passed
                    if (isset($import_ajax_input[ 'multisite_blogid_to_import' ]) && ! empty($import_ajax_input[ 'multisite_import_package_uploaded_file' ])) {
                        //Validate ID
                        $blogid_to_import	= (int) $import_ajax_input[ 'multisite_blogid_to_import' ];
                        $import_package_path = $import_ajax_input[ 'multisite_import_package_uploaded_file' ];
                        if (($blogid_to_import > 0) && (true === $this->getSystemChecks()->multisiteMigrationEssentialRequisites()) && 
                            $this->getSystemFunctions()->file_exists($import_package_path)) {
                            
                            do_action('multisite_migration_log_processed_events', 'Original handler cleared and validated, starting the import.', $blogid_to_import, 'import', 'multisiteMigrationImportProcessor', $this);
                            $all_clear	= true;
                        }
                    }
                }
            }
        }
        //Check if we are clear to proceed import processing.
        if (true === $all_clear) {
            
            //Initialize import tracker
            $this->getProgressHandlers()->initializeProgressTracker($blogid_to_import, 'import');
            
            //All clear proceed to import processing
            if (! isset($ret['diff'])) {
                $import_progress = $this->getProgressHandlers()->getTrackerProgressNonCached($blogid_to_import, 'import');
                if ($import_progress && 'boot' !== $import_progress) {
                    do_action('multisite_migration_log_processed_events', 'Import already started, aborting with progress: ' . $import_progress, $blogid_to_import, 'import', 'multisiteMigrationImportProcessor', $this);
                    wp_die();
                }                  
                $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Starting import', 'prime-mover'));
                do_action('multisite_migration_log_processed_events', 'Import request received..', 0, 'import', 'importProcessorTriggeredStatus', $this);
                if (isset($import_ajax_input)) {
                    do_action('multisite_migration_log_processed_events', $import_ajax_input, 0, 'import', 'importProcessorTriggeredStatus', $this);
                }                
            }
            
            $files_array = [];
            if (isset($import_package_path)) {
                $files_array['file'] = $import_package_path;
            }
            set_time_limit(0);
            do_action('multisite_migration_before_doing_import', $blogid_to_import);
            $results		= apply_filters('multisite_migration_do_import', $ret, $blogid_to_import, $files_array);
            
            do_action('multisite_migration_log_processed_events', 'Exited import results filter.', $blogid_to_import, 'import', 'multisiteMigrationImportProcessor', $this);
            $status = '';
            if (isset($results['diff'])) {
                //Different in system configuration is detected, alert user                
                $status = 'diffdetected';
                $nonce_to_user				= wp_create_nonce('multisite_migration_continue_import');
                $diff						= $results['diff'];                
                $diff_friendly				= $this->getSystemFunctions()->printFriendlyDiffMessages($diff);
                
                $response['diff_detected']	= true ;
                $response['diff']			= $diff_friendly;
                $response['continue_nonce']	= $nonce_to_user;
                $response['results']		= $results;       
                
                do_action('multisite_migration_log_processed_events', 'Diff detected', $blogid_to_import, 'import', 'multisiteMigrationImportProcessor', $this);
                
            } elseif ((isset($results['import_success'])) && ((true === $results['import_success']))) {
                //Import success
                do_action('multisite_migration_log_processed_events', 'Import success, returning stoptracking', $blogid_to_import, 'import', 'multisiteMigrationImportProcessor', $this);
                
                $status = 'stoptracking';
                $response['status']				= true ;
                $response['import_successful']	= '<img src="' . esc_url_raw(plugins_url('res/img/done.png', dirname(__FILE__))) . '" />';                                
            } elseif (isset($results['error'])) {
                //Import unsuccessful with specific error message
                do_action('multisite_migration_log_processed_events', 'Import error found', $blogid_to_import, 'import', 'multisiteMigrationImportProcessor', $this);
                
                $status = 'stoptracking';
                $response['status']	= false ;
                $error_message		= $results['error'];
                $response['import_not_successful'] = esc_js($error_message);
            } else {
                //Generic import error
                $response['status']= false ;
                $status = 'stoptracking';
                $response['import_not_successful'] = esc_js(__('Import failed for unknown reason.', 'prime-mover'));
            }  
            
            do_action('multisite_migration_log_processed_events', 'Saving import result output:', $blogid_to_import, 'import', 'multisiteMigrationImportProcessor', $this);
            do_action('multisite_migration_log_processed_events', $response, $blogid_to_import, 'import', 'multisiteMigrationImportProcessor', $this);
            
            $this->saveImportResultForOutput($response, $status);            
        } 
        wp_die();
    }
    
    /**
     * Save import result for output
     * @param array $response
     */
    protected function saveImportResultForOutput($response = [], $status = '') 
    {
        if ( ! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $import_id = $this->getSystemInitialization()->getImportId();
        $response['process_id'] = $import_id;
        
        $option = 'import_' . $import_id;
        $this->getSystemFunctions()->updateSiteOption($option, $response);
        if ($status) {
            $this->getProgressHandlers()->updateTrackerProgress($status);
        }
        sleep(3); 
    }
    
    /**
     * Process export request via AJAX
     * @compatible 5.6
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverSystemProcessors::itRunsPrimeMoverExportProcessorIfAllClear() 
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverSystemProcessors::itReturnsErrorIfDetectedOnExportProcessor()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverSystemProcessors::itDoesNotAllowExportProcessorCallIfNotAuthorized()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverSystemProcessors::itDoesNotAllowExportProcessorCallIfNonceFailed()
     * @tested PrimeMoverFramework\Tests\TestPrimeMoverSystemProcessors::itDoesNotAllowExportProcessorCallIfOptionInvalid()
     */
    public function multisiteMigrationExportProcessor()
    {
        if (! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $all_clear	= false;        
        $args = apply_filters( 'multisite_migration_export_processor_args' , 
            [
                'multisite_migration_export_nonce' => FILTER_SANITIZE_STRING, 
                'multisite_blogid_to_export' => FILTER_SANITIZE_NUMBER_INT, 
                'multisite_export_options' => FILTER_SANITIZE_STRING,
                'multisite_migration_export_targetid' => FILTER_SANITIZE_NUMBER_INT
            ]);

        $multisite_export_options = '';       
        
        $response	= [];        
        $export_ajax_input = $this->getSystemInitialization()->getUserInput('post', $args, 'export_processor_ajax', 'export');
        
        if (isset($export_ajax_input['multisite_migration_export_nonce'])) {
            if (wp_verify_nonce($export_ajax_input['multisite_migration_export_nonce'], 'multisite_migration_export_nonce')) {
                //Nonce passed
                if (isset($export_ajax_input[ 'multisite_blogid_to_export' ])) {
                    //Validate ID
                    $blogid_to_export	= $export_ajax_input[ 'multisite_blogid_to_export' ];
                    $blogid_to_export	= intval($blogid_to_export);
                    
                    $multisite_export_options = apply_filters('multisite_migration_get_export_option', $export_ajax_input['multisite_export_options'], $blogid_to_export);
                    if (($blogid_to_export > 0) && (true === $this->getSystemChecks()->multisiteMigrationEssentialRequisites()) && $multisite_export_options) {
                        $all_clear	= true;
                        do_action('multisite_migration_log_processed_events', 'Export parameters cleared and validated, starting the export.', $blogid_to_export, 'export', 'multisiteMigrationExportProcessor', $this);
                    }
                }
            }
        }
        if (true === $all_clear) {            
            $this->getProgressHandlers()->initializeProgressTracker($blogid_to_export, 'export');
            set_time_limit(0);
            do_action('multisite_migration_before_doing_export', $blogid_to_export);
            $this->getProgressHandlers()->updateTrackerProgress(esc_html__('Starting export', 'prime-mover'), 'export');            
            
            $export_options_array = ['multisite_export_options' => $multisite_export_options ];
            if ( ! empty($export_ajax_input['multisite_migration_export_targetid'] ) ) {
                $export_options_array['multisite_migration_export_targetid'] = $export_ajax_input['multisite_migration_export_targetid'];
            }
            
            $processor_array = apply_filters( 'multisite_migration_export_processor_array', $export_options_array, $blogid_to_export, $export_ajax_input);
            $results = apply_filters('multisite_migration_do_export', $processor_array, $blogid_to_export);
            
            do_action('multisite_migration_log_processed_events', 'Stop tracking requested', $blogid_to_export, 'export', 'multisiteMigrationExportProcessor', $this);
            do_action('multisite_migration_log_processed_events', 'Export result:', $blogid_to_export, 'export', 'multisiteMigrationExportProcessor', $this);
            do_action('multisite_migration_log_processed_events', $results, $blogid_to_export, 'export', 'multisiteMigrationExportProcessor', $this);
            
            if (isset($results['error'])) {                 
                $response['status']	= false ;
                $error_message		= $results['error'];
                $response['export_not_successful'] = esc_js($error_message); 
                do_action('multisite_migration_log_processed_events', 'Export result error found:' . $error_message, $blogid_to_export, 'export', 'multisiteMigrationExportProcessor', $this);
            } elseif ( isset($results['multisite_export_location']) && 'export_directory' === $results['multisite_export_location']) { 
                do_action('multisite_migration_log_processed_events', 'Export location defined', $blogid_to_export, 'export', 'multisiteMigrationExportProcessor', $this);
                $response['status'] = true ;
                $response['export_location'] = 'export_directory';
                $response['restore_url'] = esc_url_raw($results['download_url']);
                $response['message'] = apply_filters('multisite_migration_filter_custom_export_success_message', esc_js(__('Export saved !', 'prime-mover')), $blogid_to_export);                
              
            } elseif (isset($results['download_url'])) {  
                do_action('multisite_migration_log_processed_events', 'Download URL defined', $blogid_to_export, 'export', 'multisiteMigrationExportProcessor', $this);
                $download_url				= $results['download_url'];
                $response['status']			= true ;
                $response['download_link']	= esc_url_raw($download_url);                
            } else {
                do_action('multisite_migration_log_processed_events', 'Generic export not successful logged', $blogid_to_export, 'export', 'multisiteMigrationExportProcessor', $this);
                $response['status']= false ;
                $response['export_not_successful'] = esc_js(__('Export failed', 'prime-mover'));               
            }            
            
            do_action('multisite_migration_log_processed_events', 'Saving export result output:', $blogid_to_export, 'export', 'multisiteMigrationExportProcessor', $this);
            do_action('multisite_migration_log_processed_events', $response, $blogid_to_export, 'export', 'multisiteMigrationExportProcessor', $this);
            
            $this->saveExportResultForOutput($response, 'stoptracking');                        
        }
        wp_die();
    }

    /**
     * Save export result for outputting
     * @param array $response
     * @param string $status
     */
    protected function saveExportResultForOutput($response = [], $status = '')
    {
        if ( ! $this->getSystemAuthorization()->isUserAuthorized()) {
            return;
        }
        $export_id = $this->getSystemInitialization()->getExportId();
        $response['process_id'] = $export_id;
        
        $option = 'export_' . $export_id;
        $this->getSystemFunctions()->updateSiteOption($option, $response);
        if ($status) {
            $this->getProgressHandlers()->updateTrackerProgress($status, 'export');
        }
        sleep(3); 
    }
}
