<?php
namespace Codexonics\PrimeMoverFramework\utilities;

/*
 * This file is part of the Codexonics.PrimeMoverFramework package.
 *
 * (c) Codexonics Ltd
 *
 * This package is Open Source Software. For the full copyright and license
 * information, please view the LICENSE file which was distributed with this
 * source code.
 */

use Codexonics\PrimeMoverFramework\classes\PrimeMover;

if (! defined('ABSPATH')) {
    exit;
}

/**
 * Prime Mover Validation Utilities
 * Helper functionality for data validation processes
 *
 */
class PrimeMoverValidationUtilities
{     
    private $multisite_migration;
    private $import_utilities;
    
    /**
     * Construct
     * @param PrimeMover $multisite_migration
     * @param array $utilities
     */
    public function __construct(PrimeMover $multisite_migration, $utilities = [])
    {
        $this->multisite_migration = $multisite_migration;
        $this->import_utilities = $utilities['import_utilities'];
    }
    
    /**
     * Get multisite migration
     * @return \Codexonics\PrimeMoverFramework\classes\PrimeMover
     */
    public function getPrimeMover()
    {
        return $this->multisite_migration;
    }
    
    /**
     * 
     * Return export utilities
     */
    public function getExportUtilities()
    {
        return $this->getImportUtilities()->getExportUtilities();
    }
    
    /**
     * Get import utilities
     */
    public function getImportUtilities()
    {
        return $this->import_utilities;
    }
       
    /**
    * Initialize hooks
    */
    public function initHooks()
    {
        add_filter("multisite_migration_get_validation_id_export_processor_ajax", [$this, 'returnValidationProtocolExportProcessor'], 10); 
        add_filter("multisite_migration_get_validation_id_import_processor_ajax", [$this, 'returnValidationProtocolImportProcessor'], 10);        
        add_filter("multisite_migration_get_validation_id_diff_import_processor_ajax", [$this, 'returnValidationProtocolDiffImportProcessor'], 10);
        add_filter("multisite_migration_get_validation_id_delete_tmp_dir", [$this, 'returnValidationProtocolDeleteTmpFile'], 10);        
        add_filter("multisite_migration_get_validation_id_shutdown_error_log", [$this, 'returnValidationShutdownErrorLog'], 10);
        
        add_filter("multisite_migration_get_validation_id_download_stream_parameters", [$this, 'returnValidationDownloadStream'], 10);        
        add_filter("multisite_migration_get_validation_id_chunk_upload_ajax", [$this, 'returnValidationChunkUpload'], 10);
        add_filter("multisite_migration_get_validation_id_common_shutdown_processor", [$this, 'returnValidationCommonShutDown'], 10);
        add_filter("multisite_migration_get_validation_id_common_progress_processor", [$this, 'returnValidationCommonProgress'], 10);
        
        add_filter("multisite_migration_get_validation_id_errorlog_exist_check", [$this, 'returnValidationErrorLogExist'], 10);
        add_filter("multisite_migration_filter_error_log", [$this, 'sanitizeLog'], 10, 2);
        add_filter("multisite_migration_get_validation_id_download_authorization_parameters", [$this, 'returnValidationAuthorizationHeaders'], 10);
        add_filter("multisite_migration_get_validation_id_verify_encrypted_package", [$this, 'returnValidationProtocolDecryptVerify'], 10);
        
        add_filter("multisite_migration_get_validation_id_maintenance_control_parameters", [$this, 'returnValidationMaintenanceControlParameters'], 10);
        add_filter("multisite_migration_get_validation_id_verify_contact_page", [$this, 'returnValidationContactPageParameters'], 10);
    }

    /**
     * Return validation contact page redirect
     * @return string[][]
     */
    public function returnValidationContactPageParameters()
    {
        return [
            'page' => ['migration-panel-settings-contact']
        ];
    }
    /**
     * 
     * @return string[]|string[][]
     */
    public function returnValidationMaintenanceControlParameters()
    {
        return [
            'prime_mover_maintenance' => ['off', 'on'],
            'prime_mover_blogid' => 'positive_int'
        ];
    }
    
    /**
     * Delete tmp dir validation
     * @reviewed
     * @return string[]
     */
    public function returnValidationProtocolDecryptVerify()
    {
        return [
            'prime_mover_decryption_check_nonce' => 'nonce',
            'stringtocheck' => 'prime_mover_signature',
            'multisiteblog_id' => 'positive_int',
            'encrypted_media' => ['true', 'false']
        ];
    }
    
    /**
     * Validation protocol for authorization headers
     * @return array
     * @reviewed
     */
    public function returnValidationAuthorizationHeaders()
    {
        $current_domain = $this->getPrimeMover()->getSystemInitialization()->getDomain();
        return [
            'HTTP_X_PRIME_MOVER_DOMAIN' => [$current_domain],
            'HTTP_X_PRIME_MOVER_AUTHORIZATION' => 'sha512'
        ];
    }
    
    /**
     * Sanitize log
     * @param string $log
     */
    public function sanitizeLog($log = '', $source = '')
    {        
        if ( ! $log || ! $source ) {
            return $log;
        }
        if ($source === 'multisiteDumpDbForExport') {
            global $wpdb;
            $log = str_replace($wpdb->dbuser, 'XXXXXXX', $log);
            $log = str_replace($wpdb->dbpassword, 'XXXXXXX', $log);
            $log = str_replace($wpdb->dbhost, 'XXXXXXX', $log);
            $log = str_replace($wpdb->dbname, 'XXXXXXX', $log);
        }
        return $log;
    }

    /**
     * Get validation protocol for error log exist check
     * @return string[]
     * @reviewed
     */
    public function returnValidationErrorLogExist()
    {
        return [
            'multisite_migration_errorlog_nonce' => 'nonce',
            'error_blog_id' => 'positive_int'
        ];
    }
    
    /**
     * Return validation of commmon progress
     * @return string[]
     * @reviewed
     */
    public function returnValidationCommonProgress()
    {
        return [
            'multisite_migration_import_progress_nonce' => 'nonce',
            'multisite_migration_export_progress_nonce' => 'nonce',
            'blog_id' => 'positive_int',
            'trackercount' => 'positive_int',
            'diffmode' => 'boolean'
        ];
    }
    
    /**
     * @reviewed
     * @return string[]
     */
    public function returnValidationCommonShutDown()
    {
        return [
            'multisite_migration_import_shutdown_nonce' => 'nonce',
            'multisite_migration_export_shutdown_nonce' => 'nonce',
            'process_id' => 'sha256',
            'blog_id' => 'positive_int'
        ];        
    }
    
    /**
     * @reviewed
     * @return string[]
     */
    public function returnValidationChunkUpload()
    {
        return [
            'multisite_migration_uploads_nonce' => 'nonce',
            'multisite_blogid_to_import' => 'positive_int',
            'start' => 'float',
            'end' => 'float',
            'chunk' => 'positive_int',
            'chunks' => 'positive_int',
            'missing_chunk_to_fix' => 'positive_int'
        ];
    }
    
    /**
     * @reviewed
     * @return string[]
     */
    public function returnValidationDownloadStream()
    {
        return [
            'multisite_migration_download_nonce' => 'nonce',
            'prime_mover_export_hash' => 'sha256',
            'prime_mover_blogid' => 'positive_int',
        ];
    }
    
    /**
     * @reviewed
     * @return string[]
     */
    public function returnValidationShutdownErrorLog()
    {
        return [
            'multisite_migration_errornonce' => 'nonce',
            'prime_mover_blogid' => 'positive_int',
        ];        
    }
    
    /**
     * Delete tmp dir validation
     * @reviewed
     * @return string[]
     */
    public function returnValidationProtocolDeleteTmpFile()
    {
        return [
            'multisite_migration_deletetmpfile_nonce' => 'nonce',
            'temp_file_to_delete' => 'migration_dir',
            'diff_reject' => ['true', 'false']
        ];
    }
 
    /**
     * Diff validation protocol
     * @return string[]
     * @reviewed
     */
    public function returnValidationProtocolDiffImportProcessor()
    {
        return [
            'nonce_to_continue' => 'nonce',
            'data_to_continue' => 'diff_json',
            'diff_blog_id' => 'positive_int'
        ];
    }
    
    /**
     * Return validation protocol for export processor
     * @return string[]|string[][]
     * @reviewed
     */
    public function returnValidationProtocolImportProcessor()
    {
        return [
            'multisite_migration_import_nonce' => 'nonce',
            'multisite_blogid_to_import' => 'positive_int',
            'multisite_import_package_uploaded_file' => 'migration_package',
        ];
    }
    
    /**
    * Return validation protocol for export processor
    * @return string[]|string[][]
    * @reviewed
    */
    public function returnValidationProtocolExportProcessor()
    {
        $args = [
            'multisite_migration_export_nonce' => 'nonce',
            'multisite_blogid_to_export' => 'positive_int',
            'multisite_export_options' => $this->getExportUtilities()->getValidExportOptions(),
            'multisite_export_location' => ['export_directory', 'default'],
            'multisite_migration_encrypt_db' => ['true', 'false'],
            'multisite_migration_dropbox_upload' => ['true', 'false'],
            'multisite_migration_export_targetid' => 'positive_int'
        ]; 
        
        return $args;
    }
}